#!/usr/bin/env python
# -*- coding: utf8 -*-

#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import io
import json
from collections import defaultdict, Counter
from datetime import datetime

import pandas as pd

from fmqlutils import VISTA_DATA_BASE_DIR
from fmqlutils.cacher.cacherUtils import SCHEMA_LOCN_TEMPL 
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent
from fmqlreports.webReportUtils import SITE_DIR_TEMPL, TOP_MD_TEMPL, keyStats, roundFloat
from fmqlreports.webReportUtils import muPlotRef, makePlots, vistasOfVISNByOne

from webReportUserDirectory import csvDatesBySNO
from buildDirectory import CommonBuildDirectory
from fileDirectory import CommonFileDirectory
                
"""
Build Directory
"""

# ###################################### Totals  ############################

PLOT_PREFIX = "visn20Build"

def makeTotalsSection(cd, vistaNames, imageDir="Images", fileDirectory=None):
                       
    plotData = {}
    
    bldIds = cd.buildIds()
                       
    mu = "## Build Breakdown\n\n"
             
    countsFromTotal = {
        "__entityName": "Builds",
    
        "__total": len(bldIds),
            
        "W/ Files": len(cd.buildsWithFiles()),
        "W/ Package": len(cd.buildsWithPackage()),
        "W/ Distribution Date": len(cd.buildEntryDates())
    }
    plotData["buildPercs"] = {
            "title": f'{countsFromTotal["__total"]} Builds',
            "plotName": f'{PLOT_PREFIX}Percs',
            "plotMethod": "plotIsHasBH",
            "specs": countsFromTotal
    }
    buildFiles = cd.buildFiles()
    blurb = "The directory holds <span class='yellowIt'>{:,}</span> distinct builds in <span class='yellowIt'>{:,}</span> packages. The builds create, change or populate <span class='yellowIt'>{:,}</span> files which represents <span class='yellowIt'>{}</span> of the files in VISN 20 systems. As can be seen here, most builds specify a distribution date and are in a package but a minority contain files".format(
        countsFromTotal["__total"],
        len(cd.buildPackages()),
        len(buildFiles),
        reportPercent(len(buildFiles), len(fileDirectory.fileIds()))
    )
    mu += "{}\n\n{}\n\n".format(
        blurb, 
        muPlotRef(plotData["buildPercs"], imageDir)
    )
     
    commonTotals = defaultdict(dict) # not picturing           
    allCnt = len(cd.vistas())
    for vistaInfo in cd.vistas():
        stationNo = vistaInfo["stationNumber"]
        stationMN = vistaNames[stationNo]
        commonCountsNos = cd.commonCounts(stationNo)
        commonCountsPer = {
            "all": commonCountsNos[allCnt],
            "some": sum(commonCountsNos[i] for i in range(2, allCnt)),
            "exclusive": commonCountsNos[1]
        }
        commonTotals[stationMN] = commonCountsPer
        
    plotData["commonByVistA"] = {
            "title": "Common Build Count",
            "plotName": f'{PLOT_PREFIX}Common',
            "plotMethod": "plotCategoryBSV",
            "specs": commonTotals
    }
    blurb = "The graphic below shows the total number of builds in each system as well as a breakdown of whether those builds are unique to that system (exclusive), in some but not all systems (some) or in every (all) other systems in the directory. While the bigger systems like Puget Sound (PUG) have more builds, the vast majority of builds - <span class='yellowIt'>{}</span> - are shared by every VistA.".format(
        reportAbsAndPercent(commonTotals[list(commonTotals.keys())[0]]["all"], countsFromTotal["__total"])
    )
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(plotData["commonByVistA"], imageDir)
    )
    
    plotData["commonByVistAPerc"] = {
            "title": "Common Build Count (Percentage)",
            "plotName": f'{PLOT_PREFIX}PercCommon',
            "plotMethod": "plotCategoryBSV",
            "specs": commonTotals,
            "kargs": { 
                "usePerc": True
            }
    }
    blurb = "The following shows the same information but instead of totals, each system's representation distinguishes the percentage of its builds that are unique from those that also appear elsewhere. It highlights that bigger systems have most of the exclusive builds, that there is something distinctive about them. _These distinctions need to be investigated_."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(plotData["commonByVistAPerc"], imageDir)
    )
        
    return plotData, mu
    
# #################################### Time Series Section ##########################

def makeTimeSeriesSection(cd, vistaNames, imageDir="Images"):

    plotData = {}
    
    end = "2020-02" # so common for all (make dynamic going forward)

    mu = "## Build Additions over Time\n\n"
    
    # For now (before get installs) just doing equivalent of catch all 'user'
    # done for user report ie/ build creation dates
    dailiesDFCSV = csvDatesBySNO(
        datesBySNO={"BUILD": sorted(cd.buildEntryDates().values())}
    )
        
    colNameById = {"BUILD": "Build"}
        
    plotData["ts365RollingMean"] = {
        "title": "Trends in Build Addition (365 day rolling)",
        "plotName": f'{PLOT_PREFIX}',
        "plotMethod": "plot365RollingMean",
        "dfcsv": dailiesDFCSV,
        "kargs": {
            "colNameById": colNameById,
            "start": "1987-01", 
            "end": end
        }
    }
    blurb = "Here are the yearly trends using a __\"365 day rolling mean\"__. The build system started in 1996 and from then until 2006 produced the most builds. There was a sharp and steady decline until 2010. From that point on, build creation has been low but steady with a small blip in 2018."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(
            plotData["ts365RollingMean"],
            imageDir
        )
    )
    
    plotData["tsMonthlyResamples"] = {
        "title": "Builds - Resampled Monthly (2018 - {})".format(re.sub(r'\-', '/', end)),
        "plotName": f'{PLOT_PREFIX}',
        "plotMethod": "plotMonthlyResamples",
        "dfcsv": dailiesDFCSV,
        "kargs": {
            "colNameById": colNameById,
            "start": '2018-01', 
            "end": end
        }
    }
    blurb = "When the data is __resampled by month__, you can see the 2018 _\"blip\"_ ..."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(
            plotData["tsMonthlyResamples"],
            imageDir
        )
    )
    
    plotData["tsTotalsPerYear"] = {
        "title": "Builds - Annual Creation (2015 - {})".format(re.sub(r'\-', '/', end)),
        "plotName": f'{PLOT_PREFIX}',
        "plotMethod": "plotTotalsPerYear",
        "dfcsv": dailiesDFCSV,
        "kargs": {
            "colNameById": colNameById,
            "start": "2015", 
            "end": end,
            "snosOnly": False
        }
    }
    # TODO: calc these #'s - dangerous to just put here
    blurb = "Finally, a simple __set of totals__ for the last five years ..."
    mu += "{}\n\n{}\n\n".format(
        blurb,
        muPlotRef(
            plotData["tsTotalsPerYear"],
            imageDir
        )
    )
                                    
    return plotData, mu
                    
# ############################## DRIVER ################################
            
def main():

    assert sys.version_info >= (3, 6)
    
    if len(sys.argv) > 1:
        if sys.argv[1] == "PLOT":
            makePlots("Common", "Builds")
            return
        raise Exception("Only argument allowed is 'PLOT'")

    visn20VistAs = vistasOfVISNByOne("20")

    dirName = "Build"
    cd = CommonBuildDirectory(visn20VistAs)

    plotData = {}
    mu = f"""---
layout: default
title: VISN 20 {dirName} Directory
---

# VISN 20 {dirName} Directory

"""

    mu += """The following reports on a VISN 20 Build Directory built from the build file (_9.6_) of 7 VistAs: Puget Sound [663], Portland [648], Boise [531], Spokane [668], Roseburg [653], White City [692] and Walla Walla [687]. Together they hold <span class='yellowIt'>{:,}</span> distinct builds using <span class='yellowIt'>{:,}</span> build file entries.

Note: Anchorage is the missing VISN 20 system. Its information will be added once the system is made available.

This directory should show [1] how VistA built up over time, [2] the extent to which the build system captures the contents of VistA and how [3] common VistAs are and the components that make them different.

""".format(
        cd.total(),
        cd.entryTotal()
    )
    
    ttlPlotData, ttlMU = makeTotalsSection(cd, visn20VistAs, fileDirectory=CommonFileDirectory(visn20VistAs))
    mu += ttlMU
    plotData.update(ttlPlotData)
    
    tsPlotData, tsMU = makeTimeSeriesSection(cd, visn20VistAs)
    mu += tsMU
    plotData.update(tsPlotData)
        
    plotDir = f'{VISTA_DATA_BASE_DIR}Common/TmpWorking/'
    print(f'Serializing vizData to {plotDir}')
    json.dump(plotData, open(f'{plotDir}plotDataBuilds.json', "w"), indent=4)
                
    siteDir = SITE_DIR_TEMPL.format("Common")
    print(f'Serializing Report to {siteDir}')
    open(f'{siteDir}visn20BuildDirectory.md', "w").write(mu) 
        
if __name__ == "__main__":
    main()
    



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime, date, timedelta
from dateutil.relativedelta import relativedelta

from fmqlutils import VISTA_DATA_BASE_DIR
from fmqlutils.cacher.cacherUtils import FMQLReplyStore, FilteredResultIterator, metaOfVistA
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent, muBVC
from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, checkDataPresent, singleValue, combineSubTypes, muBVCOfSTProp

from fmqlreports.webReportUtils import TOP_MD_TEMPL, SITE_DIR_TEMPL, ensureWebReportLocations, keyStats, flattenFrequencyDistribution, roundFloat, reduce200, reduce4, flattenPropValues

"""
TODO FIRST:
- reader side report (incoming IFC ie/ filer) ... will tie in with documents <-- big for 663 and 648
- [2006.5831](http://localhost:9100/schema#2006_5831)
- may put queue trigger back in if the behavior is significant
- note if activity (ie/ if active in last year) is used in config section

Report TeleReading and the supporting IFCs, documents and images. Very particular IFC.
Cover [1] both Acquisition and Reader sides (including when they are the same VistA) and
[2] the configuration (files) used in all cases.

ADD BACK NOTE

The client __[Telereader Configurator](https://www.va.gov/vdl/documents/Clinical/Vista_Imaging_Sys/mag_telereader_configuration.pdf)__ sets up the following configuration files that determine how Telereading operates in a VistA. It configures both the __Acquisition Role__ and the __Reader Role__.

BETTER INTERP/CONFIG SETUPs (not even clear to me!):
- what's with Boise stuff? 
- remote service name vs local? ie/ one portland 'remote' "BETIC TELERETINAL IMAGING (PORTL" / 2006.5841 has many local 123.5 service types (where one each per local site like Yakama) 
     ie/ local, per instit service name [many] - global (note remote) service name
<--------- ie need to combine the configs tables ie/ local name (instit + local service name) ------> one service based remote name (per remote VistA as a whole)
- ex/ seems like 2006... Queue only uses more global sounding local service DIABETIC TELERETINAL IMAGING IFC (PORTLAND) for eye care and not all the individual per instit services supported in the system
  <------------ ie/ what does the queue use!!!
  BUT the little EYE of YAK TELE-EYE SCREENING (BOI) IFC is used ... ie/ 
- WANT TO SEE 'unused configs'!!!!!
- BIGGIE: split of 1/2/3 (local:local not IFC) ++ what's the no procedure about 
ie/ CONSISTENT configs to go with other stuff

> A (placer) consult created with the Service name (2005.5841->123.5) results in a (filler) consult with the service name Remote Name in the Remote Reading Site ... 2005.5841 also decides how entry is triggered. 

============================= TODO: Content =====================

TODO: [1] fin ac [2] finish Reader Side (align consults) + [3] HL7 vs RPC + [4] neaten/finish final rep form <------ gives new template
- the biggie is FILLER: extrapolate a list as if there is a config file for "filler reading services" ... => 
- need PUG [only internal]
- combine table better ie/ see 2006.5841 ac service -- bring in its Remote Reading Site ie/ show combination.
- RPCs/HL7 attribution and code paths (static report/table) ... see below ie/ BEYOND what data shows unless there is HL7 logs?
  - telereader use lower in SPO (only Eye) but COS consistent with all remote using it!
  - MAY REWORK USER TABLE into 'roles' based on transitions AND not the one summary in queue (who is a reader? responsible? no? ...)
- [FINAL FORM] <-------------- when done beyond freps mv, note IFC in general context + TYPER use rework as neater (not lot's of random types.) MAY DO TYPER explicitly within here ie/ don't run through front end => no need to do custom there ... build in reframe to report.

TODO WORK comments RPCs and HL7 attribution (see code etc comments in reportIFC)
  > Clinical Capture provides the capability to acquire dermatology images that can be
remotely read using Inter-Facility Consults (IFCs) and TeleReader. To support this
capability, a new association has been added to Clinical Capture called “TeleReader
Consult”. JPEG and TIFF dermatology images saved to VistA using this new association
will be converted into DICOM format before storing in VistA Imaging. The images will
then be viewable in Clinical Display.
ie/ dermatology special [and there are a bunch of them]
   ---
   > - TIU Note File – this is used to automatically create a TIU note for a set of images
when the consult request is remotely completed via an Inter-facility Consult.
ie/ TIU creation from remote IFC completion.
   > MAG3 TELEREADER READ/UNRD ADD - adds a consult and image pointers to the
UNREAD/READ LIST file (#2006.5849) and DICOM GMRC TEMP LIST file
(#2006.5839)
<------------ important: 'assoociation' of clinical capture -- find

"""

def webReportTeleReading(stationNo, stationName, onAndAfterDay, upToDay):

    allThere, details = checkDataPresent(stationNo, [

        # [ACQ] T. Q (See comment in ACQ below)
        {"fileType": "2006_5849", "check": "YR1E"},              
        # [ACQ] T. ACQUISITION SERVICE, a 'bridge' configuration file between world 
        # of 123's and VA-wide telereading. Also determines when Q entry made from 123
        # and document type used for result
        {"fileType": "2006_5841",  "check": "ALL"}, 
        
        # [READ] T. ACQUISITION SITE
        {"fileType": "2006_5842", "check": "ALL"}, 
        # [READ] T. READER 
        {"fileType": "2006_5843", "check": "ALL"}, 
        
        # REQUEST/CONSULTATION for T. Q YR1E - complication of local-only Telereading
        {"fileType": "123", "check": "YR1"}, 
        # REQUEST SERVICES (local terms mapped to VA-wide in 2006_5841)
        {"fileType": "123_5", "check": "ALL"},

        # For User Identification
        # 200 - reduce200 will enforce need for 200
        {"fileType": "3_081", "check": "YR1E"}
        
    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))

    mu = TOP_MD_TEMPL.format("{} TeleReader".format(stationNo))

    mu += """# Telereader
   
The following describes _Telereading_ in VistA _{}_ [{}]. VistA can both acquire studies (\"Acquisition\") and read studies (\"Reading\"). Where one VistA acquires a study and another reads that study, Telereading builds upon VistA's general-purpose Inter-facility Consult (IFC) feature.

    
""".format(stationName, stationNo)

    mu += webReportAcquisition(stationNo, stationName, onAndAfterDay, upToDay)
    
    mu += webReportReading(stationNo, stationName, onAndAfterDay, upToDay)
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    print("Serializing Report to {}".format(userSiteDir))
    open(userSiteDir + "telereader.md", "w").write(mu)
    
# ######################## Acquisition ######################
    
def webReportAcquisition(stationNo, stationName, onAndAfterDay, upToDay):
    mu = "## Acquisition\n\n" 
    mu += webReportAcquisitionConfiguration(stationNo, stationName)
    mu += webReportAcquisitionActivity(stationNo, onAndAfterDay, upToDay)
    return mu
    
"""
Acquisition Service is effectively a combination of information from 123.5 and 2006_5841

TODO:
- consider a synthesized name of SPEC PROC ABBREVS (SNO AC/SNO READ VISTA)
"""
def webReportAcquisitionConfiguration(stationNo, stationName):

    dataLocn = "{}{}/Data/".format(VISTA_DATA_BASE_DIR, stationNo)   
    ifcServiceIndex = dict((cs["_id"], cs) for cs in FilteredResultIterator(dataLocn, "123_5") if "ifc_routing_site" in cs)
    institByIEN = reduce4(stationNo)
    snoByNationalInstitId = dict(("4-" + ien, institByIEN[ien]["station_number"]) for ien in institByIEN if "isNational" in institByIEN[ien] and "station_number" in institByIEN[ien])
    
    asResourceIter = FilteredResultIterator(dataLocn, "2006_5841")
    serviceCount = 0
    byRoutingSite = defaultdict(list)
    locals = []
    for resource in asResourceIter:
        serviceCount += 1
        row = [
            "__{}__".format(resource["name"]["label"]),
            "{} [{}]".format( # 2006.5841
                re.sub(r'( VAMC| VA CLINIC| CBOC| VA MEDICAL CENTER)', '', resource["acquisition_site"]["label"]),
                snoByNationalInstitId[resource["acquisition_site"]["id"]]
            ),
            "{} [{}]".format( # 2006.5841
                resource["specialty_index"]["label"],
                resource["specialty_index"]["id"].split("-")[1]
            ),
            "{} [{}]".format( # 2006.5841
                resource["procedure_index"]["label"],
                resource["procedure_index"]["id"].split("-")[1]
            )
        ]
        tiuNoteFile = resource["tiu_note_file"]["label"] if "tiu_note_file" in resource else ""
        row.append(tiuNoteFile)
        # If not just LOCAL
        if resource["name"]["id"] in ifcServiceIndex: # 123.5
            ifcService = ifcServiceIndex[resource["name"]["id"]]
            row.append(ifcService["ifc_remote_name"])
            routingSite = ifcService["ifc_routing_site"]
            remoteSiteRef = "{} [{}]".format(
                routingSite["label"],
                snoByNationalInstitId[routingSite["id"]]
            )
            byRoutingSite[remoteSiteRef].append(row)
            # NOT DOING ASSOCIATED STOP CODE as many don't have it
        else:
            locals.append(row)
        # IGNORING FOR NOW (as doesn't effect quantifying per se
        # trigger = resource["unread_list_creation_trigger"]

    mu = "### Configuration\n\n"
    mu += "This VistA has <span class='countHigh'>{:,}</span> consult services for Telereading by <span class='countHigh'>{:,}</span> remote VistAs.{}\n\n".format(
        serviceCount,
        len(byRoutingSite),
        " In addition, <span class='countHigh'>{:,}</span> services are for local reading.".format(len(locals)) if len(locals) else ""
    )
    basicCols = [
            ":Service",
            ":Acquisition Site",
            ":Specialty",
            ":Procedure",
            ":Acquisition Note"
    ]
    for rs in sorted(byRoutingSite):
        # "All map to one service in Boise
        acSites = set(row[1] for row in byRoutingSite[rs])
        remoteServices = set(row[-1] for row in byRoutingSite[rs])
        cols = basicCols[:]
        if len(remoteServices) > 1:
            cols.append(":Remote Service")
        tbl = MarkdownTable(cols)
        for row in sorted(byRoutingSite[rs], key=lambda x: x[1]): # group by ac site
            if len(remoteServices) == 1:
                row.pop()
            tbl.addRow(row)
        if len(acSites) > 1:
            mu += "__{}__ receives <span class='countHigh'>{:,}</span> types of consult service from <span class='countHigh'>{:,}</span> acquisition sites{} ...\n\n".format(
                rs, 
                len(byRoutingSite[rs]), 
                len(acSites),
                ", all sent to one remote service _{}_".format(list(remoteServices)[0]) if len(remoteServices) == 1 else "" 
            )
        else:
            mu += "__{}__ receives <span class='countHigh'>{:,}</span> services ...\n\n".format(rs, len(byRoutingSite[rs]))
        mu += tbl.md() + "\n\n"
    if len(locals):
        tbl = MarkdownTable(basicCols)
        mu += "The <span class='countHigh'>{:,}</span> locally read (non interfacility consult) services are ...\n\n".format(len(locals))
        for row in sorted(locals, key=lambda x: x[0]):
            tbl.addRow(row)
        mu += tbl.md() + "\n\n"
    
    return mu

"""
Acquisition Activity - local acquisition and either remote or local telereading

Different breaks of Enhanced 2006_5849 (with a little addition from 3.081's
app breakdown)

TODO:
- more on time between consult creation (audit etc) and queue times ie/ want to get to ONE audit trail for simulation) <--------- ********
- more on local telereading (ie/ no ifc consult) <------ not in scope but ... would expect
Reading config user to be active one! [IE/ ENFORCE HERE]

The key file is Telereader List (TELEREADER READ/UNREAD LIST (2006.5849)) pre-enhanced by Placer 123's audit trail and service assertion. 

It summarized from the following perspectives and these summaries are reported on:
1. By Specialty (image_index_for_specialty)
2. By (Consult) Service (E added from 123)
3. By Status
4. By Reader (in Acquisition System's 200)

In addition, there should be a broad summary of consults as key consult props embedded
in the queue element by E.
a. consult_service_tiu_note_type (itself put into 123 from ...): would expect one as one per Consult Service
b. consult_audit_trail: into transitions mu so one value of all transitions
c. is_consult_ifc (not there if consult is local): rem E enforces that consult is 
always PLACER if ifc
e. for service: tiu note type (ie/ one or none)

number_of_images -- print it
"""
def webReportAcquisitionActivity(stationNo, onAndAfterDay, upToDay):

    # 123 and 2006_5849 reductions
    type123YR1, sts123YR1 = splitTypeDatas(stationNo, "123", reductionLabel="YR1", )
    type2006_5849YR1E, sts2006_5849YR1E = splitTypeDatas(stationNo, "2006_5849", reductionLabel="YR1E", expectSubTypeProperties=["reader_duz_at_acquisition_site", "image_index_for_specialty", "consult_service", "status", "#reading_start"])
    
    # Group per User (and don't want the one possible, no user specified)
    st2006_5849Us = [st for st in combineSubTypes(sts2006_5849YR1E, ["reader_duz_at_acquisition_site"], forceCountProps=["consult_service"]) if "reader_duz_at_acquisition_site" in st]
    readersPerAcVistA = set(singleValue(st, "reader_duz_at_acquisition_site") for st in st2006_5849Us) 
    
    # ? users == just name. What if overlap?                
    def sosOfUsers(stationNo, users):
        print("Loading 3.081 YR1E ...")
        type3_081, st3_081s = splitTypeDatas(stationNo, "3_081", reductionLabel="YR1E")    
        st3_081Us = combineSubTypes(st3_081s, ["user"], forceCountProps=["ipv4_address", "remote_station_id", "remote_user_ien", "duration"])
        st3_081ByUserRef = dict((singleValue(st, "user"), st) for st in st3_081Us if singleValue(st, "user") in readersPerAcVistA)
        print("\tLoaded and processed for {:,}".format(len(users)))
        return st3_081ByUserRef
    st3_081ByUserRef = sosOfUsers(stationNo, readersPerAcVistA) 
    
    # ? readers == just name. What if overlap?
    def autoCreatedReaders(readers):
        print("Loading user information (200 Reduction) ...")
        userInfoByUserRef = reduce200(stationNo, coreOnly=True)   
        # V1.1 and V10 compatible as creator only shows if != 200-0 for v1.1
        readersAutoCreated = set(userRef for userRef in userInfoByUserRef if userRef in readers and "creator" not in userInfoByUserRef[userRef])
        print("\tLoaded and processed {:,} users, {:,} readers to see {:,} are auto created".format(len(userInfoByUserRef), len(readers), len(readersAutoCreated)))
        return readersAutoCreated
    readersAutoCreated = autoCreatedReaders(readersPerAcVistA)
                            
    mu = "### Activity\n\n"             

    """
    Overall summary uses:
    - 123 YR1 Red for total consults
    - 2006_5849 YR1E Red for total queue entries: all have unique consult and
    with E embedding consult properties, can describe the consult of queue entries
    from the queue reduction

    # Placer Total Consults if want to use these.
        reportAbsAndPercent(
            type123YR1["ifc_role"]["byValueCount"]["P:Placer"],
            type123YR1["_total"]
        ) if "ifc_role" in type2006_5849YR1E else 0,
    """
    consultIFCCount = type2006_5849YR1E["is_consult_ifc"]["count"] if "is_consult_ifc" in type2006_5849YR1E else 0
    if consultIFCCount < type2006_5849YR1E["_total"]:
        mixedOrNotMU = "<span class='countHigh'>{}</span> of these are interfacility consults (IFCs), where another facility performs the consult while the balance, <span class='countHigh'>{}</span>, are performed locally".format(
            reportAbsAndPercent(
                consultIFCCount, 
                type2006_5849YR1E["_total"]
            ),
            reportAbsAndPercent(
                type2006_5849YR1E["_total"] - consultIFCCount, 
                type2006_5849YR1E["_total"]
            )
        )
    else:
        mixedOrNotMU = "all of which are interfacility consults (IFCs) - no locally acquired studies are read locally"
    mu += """An _Acquisition site_ imports images and submits them for telereading \"consults\". These are tracked in _TELEREADER READ/UNREAD LIST (2006.5849) [\"Acquisition List\"]_. In this site, there are <span class='countHigh'>{:,}</span> total consults between {} and {}. Of these, <span class='countHigh'>{}</span> are in the Acquisition List, {}.
    
""".format(
        type123YR1["_total"],
        onAndAfterDay, 
        upToDay,
        # count of queue consults == count of queue entries as 1-1
        reportAbsAndPercent(type2006_5849YR1E["_total"], type123YR1["_total"]), 
        mixedOrNotMU
    ) 

    """
    Specialty, Service and Status
    """
    mu += """The entries have <span class='countHigh'>{:,}</span> states, {} and <span class='countHigh'>{:,}</span> specialties, {} covering <span class='countHigh'>{:,}</span> services.
    
""".format(
        len(type2006_5849YR1E["status"]["byValueCount"]), 
        muBVC(type2006_5849YR1E["status"]["byValueCount"]), 
        len(type2006_5849YR1E["image_index_for_specialty"]["byValueCount"]), 
        muBVC(type2006_5849YR1E["image_index_for_specialty"]["byValueCount"]), 
        len(type2006_5849YR1E["consult_service"]["byValueCount"])
    )
    st2006_5849SpecialtyNService = combineSubTypes(sts2006_5849YR1E, ["image_index_for_specialty", "consult_service"], forceCountProps=["reader_duz_at_acquisition_site"]) 
    ifcServiceByIEN = reduce123_5ForIFC(stationNo)
    specialitiesTotals = Counter()
    stsBySpecialty = defaultdict(list)
    for st in st2006_5849SpecialtyNService:
        specialty = singleValue(st, "image_index_for_specialty")
        stsBySpecialty[specialty].append(st)
    for specialty in sorted(stsBySpecialty, key=lambda x: sum(st["_total"] for st in stsBySpecialty[x]), reverse=True):
        mu += "Specialty _{}_ ...\n\n".format(re.split(r' +\[', specialty)[0])
        tbl = MarkdownTable([":Service", ":Remote Reader", "Count", ":Status"])
        for stService in sorted(stsBySpecialty[specialty], key=lambda x: x["_total"], reverse=True):
            service = singleValue(stService, "consult_service")
            serviceIEN = re.search(r'\-(\d+)', service).group(1)
            name = re.split(r' +\[', service)[0]
            ifcRoutingSiteMU = re.sub(r'4\-', '', ifcServiceByIEN[serviceIEN]["ifc_routing_site"]) if serviceIEN in ifcServiceByIEN and "ifc_routing_site" in ifcServiceByIEN[serviceIEN] else ""
            tbl.addRow([
                "__{}__".format(name) if ifcRoutingSiteMU else name,
                ifcRoutingSiteMU,
                stService["_total"], 
                muBVC(stService["status"]["byValueCount"])
            ])
        mu += tbl.md() + "\n\n"
    
    """
    Status of the Queue Entries - show transition combinations for the status'. Comes
    from E on Queue [1] fabricating a consult_audit_trail_transitions property for the
    reduction and [2] having status in the breakdown.
    """
    st2006_5849Status = combineSubTypes(sts2006_5849YR1E, ["status"], forceCountProps=["consult_audit_trail_transitions"]) 
    st2006_5849ByStatus = defaultdict(list)
    for st in st2006_5849Status:
        status = singleValue(st, "status")
        st2006_5849ByStatus[status].append(st)
    # , ":Added Comments"
    tbl = MarkdownTable([":Status", ":Transitions", "Count", ":Reading Site"], includeNo=False)
    currentStatus = ""
    for status in ["W:Waiting", "U:Unread", "C:Cancelled", "L:Locked", "D:Deleted", "R:Read"]:
        if status not in st2006_5849ByStatus:
            continue
        for st in sorted(st2006_5849ByStatus[status], key=lambda x: x["_total"], reverse=True):
            # POR has this for DELETED - TODO, examine
            if "consult_audit_trail_transitions" not in st:
                continue
            for trans in sorted(st["consult_audit_trail_transitions"]["byValueCount"], key=lambda x: st["consult_audit_trail_transitions"]["byValueCount"][x], reverse=True):
                if status != currentStatus:
                    currentStatus = status
                    row = ["__{}__".format(status.split(":")[1])]
                else:
                    row = [""]
                row.extend([trans, st["consult_audit_trail_transitions"]["byValueCount"][trans]])
                if "reader_site" in st:
                    row.append(muBVC(st["reader_site"]["byValueCount"]))
                else:
                    row.append("")
                """
                ADDED COMMENT isn't in transitions - local or remote of it is
                in consult_audit_trail_comment_{} counters of local or 
                remote comments added.
                
                Would need to add 'consult_audit_trail_transitions' to typer
                and then total LOCAL or REMOTEs across a transition typer ie/
                     consult_audit_trail_comment_local - 4: 2, 5:7 ... etc  
                """
                tbl.addRow(row)
    mu += """List entries go through one or more transitions, starting with a _CPRS RELEASED ORDER_ when the entry is first added right after its consult is created. The most common sequence moves on to _RECEIVED_ (reader site received the consult) through _COMPLETE UPDATE_ (reader site has read the image and created a document) and ends with _NEW NOTE ADDED_ (an equivalent document was created in the \"local\"/acquisition site). 
    
The following shows transitions by Acquisition List entry status. The last column shows how _ADDED COMMENT_ actions accompany a specific transition. These actions may be local or remote and allow users to annotate a consult. Transitions in __bold__ represent remote reader activity ...

"""
    mu += tbl.md() + "\n\n"
    
    """
    Per Reader (User) summaries: st2006_5849Us - will distinguish remote and local 
    ... on the acquisition side ie for the 'Acquisition List'
    
    TODO: ala older report
    - network_name (so portable for common view)
    - WHERE IS THE REMOTE ID FOR ALL OF THEM!!!!
    - consider LOA as well as app count to be sure (need break!)
    
    Note: nixed roles (ie/ enterer remote or local) and remote or local tiu refs
    per user. Will return when work specific document types.
    """
    tbl = MarkdownTable([":Reader Name", "Spec(s)", "\#", ":Remote Reader", ":Applications (SOs)"])
    noLocal = 0
    for i, st in enumerate(sorted(st2006_5849Us, key=lambda x: x["_total"], reverse=True), 1):
        userRef = singleValue(st, "reader_duz_at_acquisition_site")
        readerIENPerAcVistA = re.search(r'\-(\d+)', userRef).group(1)
        appsMU = ""     
        if userRef in st3_081ByUserRef:
            st3_081 = st3_081ByUserRef[userRef]
            byAppCount = Counter() if "remote_app" not in st3_081 else st3_081["remote_app"]["byValueCount"]
            # UNSET's probably CPRS
            if st3_081["_total"] > sum(byAppCount[app] for app in byAppCount):
                byAppCount["UNSET"] = st3_081["_total"] - sum(byAppCount[app] for app in byAppCount)   
            appsMU = muBVC(byAppCount)
            for f, t in [("VISTA IMAGING TELEREADER", "__TELEREADER__"), ("MEDICAL DOMAIN WEB SERVICES", "JLV"), ("VISTAWEB-PROD", "WEB"), ("VISTA IMAGING", "IMG"), ("DISPLAY", "DISP"), ("UNSET", "_UNSET_")]:
                appsMU = re.sub(f, t, appsMU)
                
        # Note: can differ from mandatory "full_name_of_reader" (which is last reader?)
        # ... must note these going forward; readers_initials agrees with         
        # full_name_of_reader
        nameMU = "{} [{}]".format(re.sub(r'\,', ', ', userRef.split(" [")[0]), readerIENPerAcVistA)
        # Local iff one reader_duz_at_reading_site AND same as IEN at this 
        # acquisition site. Very minority thing to have (why? - see audit)
        isLocalReader = True if ("reader_duz_at_reading_site" in st and len(st["reader_duz_at_reading_site"]["byValueCount"]) == 1 and readerIENPerAcVistA == singleValue(st, "reader_duz_at_reading_site")) else False
        if isLocalReader:
            noLocal += 1
            remoteMU = ""
            if "is_consult_ifc" in st:
                print("** Don't expect Local Readers to have IFC Consults for their TeleReading")
        else:
            nameMU = "__{}__".format(nameMU)
            remoteMU = re.sub(r'VA MEDICAL CENTER', 'VAMC', muBVC(st["reader_site"]["byValueCount"], forceShowCount=False)) if "reader_site" in st else ""
            if "reader_duz_at_reading_site" in st:
                if len(st["reader_duz_at_reading_site"]["byValueCount"]) != 1:
                    remoteMU += " [AMBIG>1]"
                else:
                    remoteMU += " [{}]".format(list(st["reader_duz_at_reading_site"]["byValueCount"])[0])
        if userRef not in readersAutoCreated:
            nameMU = re.sub(r'\]', '*]', nameMU)
        tbl.addRow([
            nameMU,
            re.sub(r'(ATOLOGY| CARE)', '', muBVC(st["image_index_for_specialty"]["byValueCount"])),
            st["_total"],
            remoteMU,
            appsMU
        ])
    mu += """For the period considered here, the Acquisition List has <span class='countHigh'>{:,}</span> Readers, <span class='countHigh'>{}</span> of which are local. The names of Readers from remote sites (_\"Remote Readers\"_) are in __bold__.
    
""".format(
        len(st2006_5849Us), 
        reportAbsAndPercent(noLocal, len(st2006_5849Us))
    )
    mu += """<span class='countHigh'>{}</span> of the following Readers were added automatically to this VistA - those added explicitly are marked with a _*_. \"Applications (Signons)\" gives the number of sign ons a user makes with different applications for the period under consideration. _UNSET_ means the application doesn't identify itself - _CPRS_ is the chief example of this. Note that _TeleReader_ is used by remote users.
    
""".format(
        reportAbsAndPercent(
            len(readersAutoCreated),
            len(st2006_5849Us)
        )
    )
    mu += tbl.md() + "\n\n"
        
    return mu  
        
# ################## Configuration (Reader and Acquisition Roles) ###############
    
def webReportReading(stationNo, stationName, onAndAfterDay, upToDay):
    mu = "## Reading\n\n" 
    mu += webReportReadingConfiguration(stationNo, stationName)
    # mu += webReportReadingActivity(stationNo, onAndAfterDay, upToDay)
    return mu
    
def webReportReadingConfiguration(stationNo, stationName):
        
    # > The TELEREADER ACQUISITION SITE file (#2006.5842) contains the list of the
    # Acquisition Sites for each Reading Site. This list must also include the Reading     
    # Site itself if it is to perform image acquisition as well.
    # > ...
    def mu2006_5842(dataLocn):

        propMap = [("name", "Name", "DROP"), ("primary_site", "Primary Site", "IEN"), ("lock_timeout_in_minutes", "Lock Timeout")] # dropped status
        
        resourceIter = FilteredResultIterator(dataLocn, "2006_5842")
        tbl = MarkdownTable([":{}".format(pmap[1]) for pmap in propMap])
        includesItselfMU = ""
        for i, resource in enumerate(sorted(resourceIter, key=lambda x: x["primary_site"]["label"], reverse=True), 1):
            resource["name"]["label"] = re.sub(r'( VAMC| CBOC| VA CLINIC| VA MEDICAL CENTER)', "", resource["name"]["label"])
            flattenPropValues(resource)
            if re.search(stationNo, resource["primary_site"]):
                includesItselfMU = ", including itself"
            tbl.addRow(rowUp(resource, propMap, True, suppress=["status"]))

        fmu = """Despite the name, _TELEREADER ACQUISITION SITE [2006.5842]_ configures a VistA for __Reading__, not acquisition. It lists __the sites a VistA can read from__. This list will include the Reading Site itself if it reads images acquired locally. In this VistA, it has <span class='countHigh'>{:,}</span> entries{} ... 

""".format(i, includesItselfMU)
        fmu += tbl.md() + "\n\n"
        return fmu
        
    # Add more 200 details ie/ created explicitly or not? Presume so
    # Add -- should tie into FILLER Consults?
    def mu2006_5843(dataLocn):
    
        propMap = [("reader", "Reader"), ("acquisition_site", "Acquisition Site", "", "acquisition_site")]
        
        resourceIter = FilteredResultIterator(dataLocn, "2006_5843")
        tbl = MarkdownTable([":{}".format(pmap[1]) for pmap in propMap])
        for i, resource in enumerate(resourceIter, 1):
            flattenPropValues(resource)
            tbl.addRow(rowUp(resource, propMap, True))
    
        fmu = """_TELEREADER READER [2006.5843]_ adds information about individual users for __Reading__. It [1] identifies readers using their User file entry (200), [2] lists the  one or more _TELEREADER ACQUISITION SITE [2006.5842]_ the user can read from and [3] designates the clinical specialties a user has. In this VistA, it has <span class='countHigh'>{:,}</span> entries ... 

""".format(i)
        fmu += tbl.md() + "\n\n"
        return fmu
        
    dataLocn = "{}{}/Data/".format(VISTA_DATA_BASE_DIR, stationNo)   

    mu = "### Configuration\n\n"
    mu += mu2006_5842(dataLocn)
    mu += mu2006_5843(dataLocn)
    
    return mu
        
# ################################# Utils #############################
    
def rowUp(resource, propMap, enforceMapAll=True, suppress=[]):
    propsDontSee = ["type", "_id", "label"]
    propsDontSee.extend(suppress)
    propsSeen = set(k for k in resource if k not in propsDontSee)
    propsMapped = set(suppress)
    row = []
    for i, pMap in enumerate(propMap, 1):
        if pMap[0] in resource:
            propsMapped.add(pMap[0])
            if len(pMap) > 3 and pMap[3]:
                valMU = ", ".join(sorted([re.sub(r' +\[[\d\_]+\-', ' [', v[pMap[3]]) for v in resource[pMap[0]]]))
                row.append(valMU)
                continue
            val = resource[pMap[0]]
            if len(pMap) == 3:
                if pMap[2] == "IEN":
                    val = re.sub(r' +\[[\d\_]+\-', ' [', val)
                elif pMap[2] == "DROP":
                    val = re.split(r' +\[', val)[0]
            valMU = "__{}__".format(re.sub(r'\s+$', '', val)) if i == 1 else val
            row.append(valMU)
        else:
            row.append("")
    if enforceMapAll and len(propsSeen - propsMapped):
        raise Exception("Not all resource props expected were mapped")
    return row
    
# ############### Reducer/Reframer to nix for Typers #############
    
"""
TODO: part of 123 E, accounting for 2006_5849. REM: each 2006_5849 has one and only
one 2006_5849 => can use consult totals to tell how many missing.

REWORK COMPLETELY to do typer in MU of 123.

Consults have a service (123_5). If a service appears in TELEREADER ACQUISITION SERVICE (2006_5841) then it is used for Acquisition. However, some consults DO NOT APPEAR in
the Queue. This isolates them.

TODO: part of IFC in general and part of 123E walk in general.

Note: this is the direction from Consult to Queue. This COULD become type based where
123 E to feature whether it is in 2006.5849. See db123DetailsByIEN


    TODO: will need 123 reduction where its #'s exceed those of 2006_5849.
    
    Report this misconfiguration of the Acquisition Service List
    ie/ consults expected to be in Queue are not there (rightly so). The Acquisition 
    Service List is wrong. Two reasons seen:
    - Service is actually for reading for a remote site
    - Service is a Procedure (and therefore won't go in TeleReading?)
    - 668 has two others, dental, may be turned on for experiment and deleted?
    
    isolateMissingConsults <---- replace with MU direct of 123E where 123E
    has more explicit markers to count the missing's.

    if "missingFromQueue" in trqConsultSummary:
        # This means service SHOULD NOT be defined for Acquisition as it 
        # is used by a remote site to send consults to this site for Reading!
        missingMU = "Acquisition configuration predicts <span class='countHigh'>{:,}</span> consults should be on the LIST but are not.".format(trqConsultSummary["missingFromQueue"])
        if "missingFromQueueFillers" in trqConsultSummary:
            missingMU += " <span class='countHigh'>{:,}</span> belong to service(s) _{}_ that are Readers for remote VistAs.".format(
                trqConsultSummary["missingFromQueueFillers"],
                ", ".join(trqConsultSummary["missingFromQueueFillersServices"])
            )
        if "missingFromQueueProcedure" in trqConsultSummary:
            missingMU += " <span class='countHigh'>{:,}</span> belong to service(s) _{}_ which involve procedures.".format(
                trqConsultSummary["missingFromQueueProcedure"],
                ", ".join(trqConsultSummary["missingFromQueueProcedureServices"])
            )           
        if "missingFromQueueLocalDiscontinue2" in trqConsultSummary:
            missingMU += " <span class='countHigh'>{:,}</span> are discontinued before they reach the queue.".format(
                trqConsultSummary["missingFromQueueLocalDiscontinue2"]
            )           
        if "otherMissing" in trqConsultSummary:
            missingMU += " <span class='countHigh'>{:,}</span> are absent for other reasons.".format(
                trqConsultSummary["otherMissing"]
            )
"""
def isolateMissingConsults(stationNo): # TMP - move to use typer in mu

    csJSNFl = "{}{}/TmpWorking/".format(VISTA_DATA_BASE_DIR, stationNo) + "telereaderQueueConsultSummary.json"
    try:
        trqConsultSummary = json.load(open(csJSNFl))
    except:
        pass
    else:
        return trqConsultSummary
        
    raise Exception("TODO: need queueConsultIds and consultRedById")

    # TODO: replace consultRedById with walk of 123's in time period 
    trqConsultSummary = {
        "total": len(consultRedById),
    }        
    acquisitionServiceIds = set(id for id in consultRedById if "is_service_for_acquisition" in consultRedById[id])
    missingFromQueue = acquisitionServiceIds - queueConsultIds
    if len(missingFromQueue):
        trqConsultSummary["missingFromQueue"] = len(missingFromQueue)
        missingFromQueueFillers = set(consultId for consultId in missingFromQueue if "ifc_role" in consultRedById[consultId] and consultRedById[consultId]["ifc_role"] == "F:FILLER")
        if len(missingFromQueueFillers):
            trqConsultSummary["missingFromQueueFillers"] = len(missingFromQueueFillers)
            missingFromQueueFillersServices = set(consultRedById[consultId]["to_service"] for consultId in missingFromQueueFillers)
            trqConsultSummary["missingFromQueueFillersServices"] = sorted(list(missingFromQueueFillersServices))
        missingFromQueueProcedure = set(consultId for consultId in missingFromQueue if "is_acquisition_service_procedure" in consultRedById[consultId])
        if len(missingFromQueueProcedure):
            trqConsultSummary["missingFromQueueProcedure"] = len(missingFromQueueProcedure)
            missingFromQueueProcedureServices = set(consultRedById[consultId]["to_service"] for consultId in missingFromQueueProcedure) 
            trqConsultSummary["missingFromQueueProcedureServices"] = sorted(list(missingFromQueueProcedureServices))         
        missingFromQueueLocalDiscontinue2 = set(consultId for consultId in missingFromQueue if "ifc_role" in consultRedById[consultId] and consultRedById[consultId]["ifc_role"] == "P:PLACER" and re.match(r'DISCONTINUED', consultRedById[consultId]["cprs_status"]))
        if len(missingFromQueueLocalDiscontinue2):
            trqConsultSummary["missingFromQueueLocalDiscontinue2"] = len(missingFromQueueLocalDiscontinue2)
        otherMissing = missingFromQueue - ((missingFromQueueFillers.union(missingFromQueueProcedure)).union(missingFromQueueLocalDiscontinue2))
        if len(otherMissing):
            trqConsultSummary["otherMissing"] = len(otherMissing)
    json.dump(trqConsultSummary, open(csJSNFl, "w"))        
    return trqConsultSummary
        
"""            
    Combination of 123.5 (has or hasn't 'Remote Reading Site') and 2006.5841 
    (identifies specifically as telereader acquisition service) identifies telereading
    location and the shape of Consults. 
    
    For all telereader acquistion services (in 2006.5841) ...
    
            Read     123.5 Has 'Remote Reading Site'    123 Role
            ----        -------------------            ----------
            LOCAL               NO                        NONE
            IFC/REMOTE          YES                     P:PLACER
            
    but there's room for error in configuration. In 757, one LOCAL site
    (acquire and read on-site locally) is never used for local reading of
    local acquisitions (ifc_role NONE). All of its consults have ifc_role
    F:FILLER which points to it being a remote reader for another site, one
    that should NOT be in 2006.5841. This can only be known by examining the
    consults of a service and excluding services from acquisition that are
    F:FILLER.
    
    Note that so far, don't see any mixed situations ie/ remote site treats
    a Local Service as its remote Reader AND the site does so itself too ie.
    2006.5841 is valid but the service is doing double duty.
"""
def reduce123_5ForIFC(stationNo):
    
    jsnFl = "{}{}/TmpWorking/".format(VISTA_DATA_BASE_DIR, stationNo) + "ifcServiceByIEN.json"
    try:
        ifcServiceByIEN = json.load(open(jsnFl))
    except:
        pass
    else:
        return ifcServiceByIEN
    
    dataLocn = "{}{}/Data/".format(VISTA_DATA_BASE_DIR, stationNo)   
    resourceIter = FilteredResultIterator(dataLocn, "123_5")
    ifcServiceByIEN = {}
    for i, resource in enumerate(resourceIter, 1):
        ien = resource["_id"].split("-")[1]
        if not ("ifc_remote_name" in resource or "ifc_routing_site" in resource):
            continue
        flattenPropValues(resource)
        ifcServiceByIEN[ien] = {"service_name": resource["service_name"]}
        if "ifc_remote_name" in resource:
            ifcServiceByIEN[ien]["ifc_remote_name"] = resource["ifc_remote_name"]
        if "ifc_routing_site" in resource:
            ifcServiceByIEN[ien]["ifc_routing_site"] = resource["ifc_routing_site"]
        if "associated_stop_code" in resource:
            ifcServiceByIEN[ien]["associated_stop_code"] = list(set(sr["associated_stop_code"] for sr in resource["associated_stop_code"]))
        
    json.dump(ifcServiceByIEN, open(jsnFl, "w"))
        
    return ifcServiceByIEN  
                        
# ################################# DRIVER #######################
               
def main():

    assert sys.version_info >= (3, 6)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO")

    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise SystemExit("Expect User Site to already exist with its basic contents")

    # <=> YR1 (fits with YR1E of 3.081)
    meta = metaOfVistA(stationNo)
    vistaName = "VistA" if "name" not in meta else meta["name"]
    cutDate = meta["cutDate"]
    cutDateDT = datetime.strptime(cutDate.split("T")[0], "%Y-%m-%d")
    yrBeforeDateDT = cutDateDT - relativedelta(years=1)
    yrBeforeDate = datetime.strftime(yrBeforeDateDT, "%Y-%m-%d")
    onAndAfterDay = yrBeforeDate
    upToDay = cutDate
    print("Reporting TeleReading for YR1 (from {} to {})".format(onAndAfterDay, upToDay))

    webReportTeleReading(stationNo, meta.get("name", "VistA"), onAndAfterDay, upToDay)
        
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta

from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent, muBVC
from fmqlutils.cacher.cacherUtils import metaOfVistA, TMPWORKING_LOCN_TEMPL, FMQLReplyStore, FilteredResultIterator, DATA_LOCN_TEMPL 
from fmqlutils.typer.reduceTypeUtils import checkDataPresent, muBVCOfSTProp

from ..webReportUtils import TOP_MD_TEMPL, SITE_DIR_TEMPL, ensureWebReportLocations, keyStats, roundFloat, reduce44, filteredResultIterator, flattenPropValues
        
"""  
Encounter (1 or more visits) matched to Appointment => Scheduled Encounter; Appointment matched to Encounter. [1] How well does the system do it? [2] And what's the institution and stop code breakdown. 

TODO/Evolve: [1] break down between two roles - [a] matching by system (how good?) and [b] then of the matched, the institution/stop code breakdown ie/ data problems and clean
data reporting (** there are encounters from cancelled and no show appts - can you tell from the visit?; are some non scheduled encounters really just same day extensions of scheduled encounters?**) [2] move to simpler indexes (apptN to Enc etc) for use in reportScheduling
and in a rewrite of the below to use typer. [3] for each work type broadly (DENTAL etc), do narrow report

TODO: Weekdays only n Nature of Clinics
- ex/ YAKIMA CBOC [687] - 687HA in FOIA - M-F (https://www.va.gov/directory/guide/facility.asp?id=5319)  
  < The clinic provides primary medical and mental health care with full-time staff.  The clinic contracts locally in the community and surrounding area for emergent laboratory, x-ray and pharmacy services.
- ex/ Coeur d' Alene
(https://www.va.gov/directory/guide/facility.asp?id=5674)

ISSUE/BUG: 541 - too big in first reduction so removed _06 pass. Could break this pass into four.

NICETY TODO: PIE chart the Instit table ie/ those columns into obvious visual break outs

End Goal: explicit RF'ed data makes Appointment --> Encounter --> Activity clear
where locations too are clear. ie/ where, when, what type of work and scheduled
or not and get away from fact systems making rawer data are not properly synced
even where these (sub)systems are modules of VistA (VPE to IHS Visits etc)
ie/ MVDM (Encounter, Appointment) from VDM (Visit, 409_84, 44_003, V PROVIDER)

cgfmr/issues:114 [CONTEXT: need to get to replay]
- rerun gatherEncounterDataByDay for 668 and 663 (663 no child visits -- why?)
"""
    
"""
webReportEncounters (appts and encounters, scheduled or not)

Mixes QA (mismatch and match of enc and appt in system and by it) with per site
and worktype breakdowns. TODO: make this distinction clearer
"""
def webReportEncounters(stationNo, vistaName, indexedDayTimeLocationPatient, cutDate, onAndAfterDay, upToDay, printOnly=True):

    byCatag = Counter()
    noneApptOnlyDetails = Counter()
    matchedDetails = {"encounterTypes": Counter(), "apptTypeByCatag": defaultdict(lambda: Counter())}
    matchedDetails2ByCatag = defaultdict(lambda: Counter())
    visitCreateBorAStart = defaultdict(lambda: Counter())
            
    """
    TODO: move to custom + typer and off the embedded below
        
    For analysis, custom gathers and adds new props to visit
    
    New/Moved Props:
        - apptCatag 
        - _children (move)
        - secondary credit stop
        - _isNoShow
        - _isCancelled
        - _isWalkin
        (ie/ some from visit gather/process, some from appt prop transfer)
        
    Config typer, can make all these =>  
        - cross: ALL
        - byInstit
        - byInstit, dssRef
        - byInstit, createOption
        - byInstit, location
        => break by all four (built in so fine)
        
    Note: tried service_category, encounter_type, #parent_visit_link where expect
    et=CREDIT STOP to be for parent_visit_link only (subset of them)
    """ 
    
    totalInfos = 0 # ie/ d/t/l/p
    totalSlots = 0
    totalMatchedAppts = 0
    catagedEncounters = [] # post 'custom' so have props
    unmatchedAppts = [] # pick out the active one
    
    hlInfosByIEN = reduce44(stationNo) # to get instit of locn for unmatched appts
    hlInfosByLabel = defaultdict(list)
    for ien in hlInfosByIEN:
        hlInfo = hlInfosByIEN[ien]
        label = hlInfo["label"]
        hlInfosByLabel[label].append(hlInfo)
    # Let apptN superceed
    def matchOperativeAppointment(info):
        for defn in [
            ("apptN", "", "active"), 
            ("apptNCancelled", "_isCancelled", "cancelled"), 
            ("apptNNoShow", "_isNoShow", "noShow"), 
            ("apptL", "", "active"),
            ("apptLC", "_isCancelled", "cancelled")
        ]:
            if defn[0] in info:
                return defn, info[defn[0]][0]
        raise Exception("Expect to match some operative appointment")
    for day in indexedDayTimeLocationPatient:
        for tm in indexedDayTimeLocationPatient[day]:
            totalSlots += 1
            for locn in indexedDayTimeLocationPatient[day][tm]:
                for patient in indexedDayTimeLocationPatient[day][tm][locn]:
                    info = indexedDayTimeLocationPatient[day][tm][locn][patient]
                    totalInfos += 1
                    try:
                        catag = categorizeVisitAppts(info, noneApptOnlyDetails, matchedDetails, matchedDetails2ByCatag, visitCreateBorAStart)            
                    except:
                        raise
                    byCatag[catag] += 1
                    if catag == "NONE_PCEONLY":
                        visitNS = info["visitNS"][0]
                        visitNS["_apptCatag"] = catag
                        catagedEncounters.append(visitNS)
                        continue
                    # No appts
                    if catag == "NONE_SCHEDONLY":
                        visit = info["visit"][0]
                        visit["_apptCatag"] = catag
                        catagedEncounters.append(visit)  
                        continue                      
                    if re.match(r'SCHED', catag):
                        visit = info["visit"][0]
                        visit["_apptCatag"] = catag
                        catagedEncounters.append(visit)
                        totalMatchedAppts += 1
                        apptSpec, appt = matchOperativeAppointment(info)
                        if apptSpec[1]:
                            visit[apptSpec[1]] = True
                        continue
                    if re.match(r'PCE', catag):
                        visitNS = info["visitNS"][0]
                        visitNS["_apptCatag"] = catag
                        catagedEncounters.append(visitNS)
                        totalMatchedAppts += 1
                        apptSpec, appt = matchOperativeAppointment(info)
                        if apptSpec[1]:
                            visitNS[apptSpec[1]] = True
                        continue
                    if catag == "NONE_NOTMATCHABLE":
                        if not ("ehVisit" in info and len(info) == 1): 
                            raise Exception("HUH on NONE")
                        """ 
                        # COME BACK
                        if "visit" in info:
                            visit = info["visit"][0]
                        else:
                            visit = info["visitNS"][0]
                        visit["_apptCatag"] = catag
                        catagedEncounters.append(visit)
                        """
                        continue
                    # NONE_APPTONLY - only one NOT visit based (most cancelled)
                    if catag == "NONE_APPTONLY":
                        apptSpec, appt = matchOperativeAppointment(info)
                        appt["_location"] = locn
                        if locn in hlInfosByLabel:
                            hlInfos = hlInfosByLabel[locn]
                            if len(hlInfos) > 1:
                                for hlInfo in hlInfos:
                                    if "institution" in hlInfo:
                                        break
                                print("** Warning - ambig label {} for locations. Just taking institution from first".format(locn))
                            else:
                                hlInfo = hlInfos[0]
                            if "institution" in hlInfo:
                                appt["_institution"] = hlInfo["institution"]
                        if apptSpec[1]:
                            appt[apptSpec[1]] = True
                        unmatchedAppts.append(appt)
                        continue
                    raise Exception("Unexpected catag: {}".format(catag))
                    
    # Will use catagedEncounters, unmatchedAppts below
    print("Post Walk of {:,} Slots in {:,} D/T/L/P's, have {:,} cataged visits, {:,} matched appts and {:,} unmatched appts. Specifically {} unmatched appts are cancelled, {} are noshows.".format(
        totalSlots,
        totalInfos,
        len(catagedEncounters),
        totalMatchedAppts,
        len(unmatchedAppts),
        
        reportAbsAndPercent(sum(1 for appt in unmatchedAppts if "_isCancelled" in appt), len(unmatchedAppts)),
        reportAbsAndPercent(sum(1 for appt in unmatchedAppts if "_isNoShow" in appt), len(unmatchedAppts))
    ))
        
    # Expect vast majority to be cancelled
    def summarizeUnmatchedAppts(appts):
        
        summ = {
            "id": "unmatchedAppts",
            "total": 0,
            "patients": set(),
            "locns": set(),
            "active": 0,
            "cancel": 0,
            "noShow": 0
        }
        summByInstit = {}
        for appt in appts:
            if "_institution" not in appt: # skipping if couldn't work out above!
                continue
            instit = appt["_institution"].split(" [")[0]
            if instit not in summByInstit:
                byInstit = {
                    "id": "unmatchedAppts{}".format(instit),
                    "total": 0,
                    "patients": set(),
                    "locns": set(),
                    "active": 0,
                    "cancel": 0,
                    "noShow": 0
                }
                summByInstit[instit] = byInstit
            else:
                byInstit = summByInstit[instit]
            for s in [summ, byInstit]:
                s["total"] += 1
                s["patients"].add(re.search(r'\-(\d+)', appt["patient"]).group(1))
                s["locns"].add(re.sub(r'\_', '/', appt["_location"]))
                if "_isCancelled" in appt:
                    s["cancel"] += 1
                elif "_isNoShow" in appt:
                    s["noShow"] += 1
                else:
                    s["active"] += 1 
        return summ, summByInstit
                    
    """
    TODO: replace with typer 4 as it will count appropriately.
    """            
    def summarizeAllEncounters(visits):    
                  
        # add to utils with makeSummary          
        def summarizeEncounters(visits, totalersAndAdders=[]):
        
            # TMP: TODO move into visit to encounter reduction code
            # WARNING in 541: location WMIC, WSIC only - seems to be all IN patient, TIU
            # from CPRS ie/ part of long stay. But not common.
            def enforceChild(visit): # bonus - move this
                if "_child_visits" in visit:
                    for cvisit in visit["_child_visits"]:
                        if cvisit["service_category"] not in ["E:EVENT (HISTORICAL)", "D:DAILY HOSPITALIZATION DATA", "X:ANCILLARY PACKAGE DAILY DATA"]: 
                            print("** WARNING: Children only expected to be EH or D:DAILY or X:ANCILL - {} is service_category of child with location {} and status {}".format(cvisit["service_category"], cvisit["hospital_location"].split(" [")[0] if "hospital_location" in cvisit else "", cvisit["patient_status_in_out"] if "patient_status_in_out" in cvisit else ""))
        
            for visit in visits:
            
                enforceChild(visit)
                
                if visit["_apptCatag"] == "NONE_NOTMATCHABLE":
                    if "service_category" in visit and re.match(r'E:EVENT', visit["service_category"]):
                        taa["ehVisit"] += 1 # DEBUG as not working
                    continue
            
                for taa in totalersAndAdders:

                    taa["total"] += 1

                    taa["patients"].add(re.search(r'\-(\d+)', visit["patient_name"]).group(1))
                                        
                    if "_child_visits" in visit:
                        taa["haveChildren"] += 1
                        taa["totalChildren"] += 1
                    
                    # TODO? not accounting for no locn
                    if "hospital_location" in visit:
                        taa["locns"].add(re.sub(r'\_', '/', visit["hospital_location"].split(" [")[0]))
                        
                    if "createdBys" in taa:
                        userRef = visit["created_by_user"] if "created_by_user" in visit else "*NOT SPECIFIED*"
                        taa["createdBys"].add(userRef)
                                                
                    if visit["patient_status_in_out"] == "1:IN":
                        taa["noIns"] += 1
                        """
                        REMOVING as mainly true but not always
                        if not ("service_category" in visit and visit["service_category"] in ["D:DAILY HOSPITALIZATION DATA", "I:IN HOSPITAL"]):
                            print(json.dumps(visit, indent=4))
                            raise Exception("Expected sc to be hosp related for in patient visit")
                        """

                    if "byServiceCategory" in taa:
                        sc = visit["service_category"] if "service_category" in visit else "*NOT SPECIFIED*"
                        et = visit["encounter_type"] if "encounter_type" in visit else "*NOT SPECIFIED*"
                        oc = visit["option_used_to_create"] if "option_used_to_create" in visit else "*NOT SPECIFIED*"
                        taa["byServiceCategory"][sc] += 1
                        taa["byEncounterType"][et] += 1
                        taa["byOptionCreateUsed"][oc] += 1
                            
                    if re.match(r'SCHED', visit["_apptCatag"]):
                        taa["schedAuto"] += 1
                        if "_isCancelled" in visit:
                            taa["schedAutoCancel"] += 1
                        elif "_isNoShow" in visit:
                            taa["schedAutoNoShow"] += 1
                        else:
                            taa["schedAutoActive"] += 1
                        continue
                    if re.match(r'PCE', visit["_apptCatag"]):
                        taa["schedManual"] += 1
                        if "_isCancelled" in visit:
                            taa["schedManualCancel"] += 1  
                        elif "_isNoShow" in visit:
                            taa["schedManualNoShow"] += 1
                        else:
                            taa["schedManualActive"] += 1
                        continue
                    if visit["_apptCatag"] == "NONE_PCEONLY":
                        taa["pceOnly"] += 1
                        continue
                    if visit["_apptCatag"] != "NONE_SCHEDONLY":
                        raise Exception("Expected NONE SCHED ONLY: {}".format(visit["_apptCatag"]))
                    taa["schedOnly"] += 1
        
        """
        TODO TYPER REPLACE THESE SUMMARIES:
        - <=> ["dss_id", "loc_of_encounter"] BUT what if loc_of_encounter missing
        
        - cross <=> all with locations, patients forced to keep
        - summByInstit <=> stInstitByInstit
        - dssSummsByInstit <=> stDSSInstitByInstit ie/ broken into DSS's
        """  
        encountersByInstit = defaultdict(list)
        for visit in visits:
            if "loc_of_encounter" not in visit:
                raise Exception("Expected all Encounter visits to have location of encounter specified")
            instit = visit["loc_of_encounter"].split(" [")[0]
            encountersByInstit[instit].append(visit)
            
        """
        Note: no walkin and that's available for apptN. Could add later (would subset actives)
        """
        def makeSummary(id):
            return {
                "id": id,
                    
                "locns": set(),
                "patients": set(),
                "createdBys": set(), # PREP for machine users or not
            
                "total": 0,
                "noIns": 0,
                
                "haveChildren": 0,
                "totalChildren": 0,
                    
                "schedAuto": 0,
                "schedAutoActive": 0, # ie/ active non walkin
                "schedAutoNoShow": 0,
                "schedAutoCancel": 0,
                "schedManual": 0,
                "schedManualActive": 0,
                "schedManualNoShow": 0,
                "schedManualCancel": 0,
                "schedOnly": 0,
                
                "pceOnly": 0,
                    
                "ehVisit": 0, # NOTE: NOT INCLUDED IN TOTALS ABOVE including 'total' ... stand alone for later. See "COM_CARE" in Cleveland so TODO undercounting
                    
                # Want these for non event hist only
                "byServiceCategory": Counter(),
                "byEncounterType": Counter(),
                "byOptionCreateUsed": Counter()   
            }
                          
        cross = makeSummary("cross")
        
        summByInstit = {}
        dssSummsByInstit = defaultdict(list)
        createOptionSummsByInstit = defaultdict(list)
        
        noOptionCreateBy = Counter()
                
        for instit in encountersByInstit:
                                    
            byInstit = makeSummary(instit)
            summByInstit[instit] = byInstit
            
            summarizeEncounters(encountersByInstit[instit], [byInstit, cross])

            visitsByDSSRef = defaultdict(list)
            visitsByCreateOption = defaultdict(list)
            for visit in encountersByInstit[instit]:
                if "dss_id" not in visit:
                    dssRef = "*NOT SPECIFIED*"
                else:
                    dssRef = visit["dss_id"]
                visitsByDSSRef[dssRef].append(visit)
                if "option_used_to_create" not in visit:
                    createOptionRef = "*NOT SPECIFIED*"
                else:
                    createOptionRef = visit["option_used_to_create"]
                visitsByCreateOption[createOptionRef].append(visit)

            for dssRef in visitsByDSSRef:
             
                byDSS = makeSummary(dssRef)
                
                dssSummsByInstit[instit].append(byDSS)

                summarizeEncounters(visitsByDSSRef[dssRef], [byDSS])
                
            for createOption in visitsByCreateOption:
             
                byCreateOption = makeSummary(createOption)
                
                createOptionSummsByInstit[instit].append(byCreateOption)

                summarizeEncounters(visitsByCreateOption[createOption], [byCreateOption])
                
            # TODO: visits by location to just focus on that in detail if need be
                
        return cross, summByInstit, dssSummsByInstit, createOptionSummsByInstit
        
    unmatchedApptsSumm, unmatchedApptsSummByInstit = summarizeUnmatchedAppts(unmatchedAppts)
        
    # Will all go to different typer passes/summary reductions
    cross, summByInstit, dssSummsByInstit, createOptionSummsByInstit = summarizeAllEncounters(catagedEncounters)
    
    # Appt or Encounter 
    allPatients = cross["patients"].union(unmatchedApptsSumm["patients"])
    print("** Warning: unmatched appointments introduce {:,} novel patients".format(len(unmatchedApptsSumm["patients"] - cross["patients"])))
    allLocns = cross["locns"].union(unmatchedApptsSumm["locns"])
    print("** Warning: unmatched appointments introduce {:,} novel locations".format(len(unmatchedApptsSumm["locns"] - cross["locns"])))
    
    # ############################ MU .md #####################################
    # TODO: add MU excel
                 
    def percentBreakdownMU(counts, contextTotal=-1, bold=True): 
        total = sum(cnt for cnt in counts)
        if total == 0:
            return "-"
        if contextTotal == -1:
            contextTotal = total
        percents = [re.sub(r'^0$', '-', re.sub(r'\.0$', '', re.sub(r'\%', '', reportPercent(cnt, contextTotal)))) for cnt in counts]
        mu = "/".join(percents)
        return "__{}__".format(mu) if bold and float(total)/float(contextTotal) > 0.5 else mu
                 
    def muByInstit(cross, summByInstit, unmatchedApptsSumm, unmatchedApptsSummByInstit):
        mu = """The following quantifies both appointments and encounters by institution.
  * _Encounters_ shows the quantity and percentage of encounters in each institution for this period. Note that this number does not include dependent (\"child\") or historical encounters. 
  * _Sched_ quantifies encounters with matching appointments (\"Scheduled Encounters\").  The _AU/M/O%_ breakdown shows how many of these are _AUtomatically_ entered into VistA as a result of appointment creation, how many were _Manually_ entered, prompted by an appointment and how many are marked as scheduled by VistA but which lack an appointment. The latter reflect synchronization bugs in the system. Its entry is in _Bold_ below when  _Scheduled Encounters_ are in the majority.
  * _NoSched_ quantifies encounters without appointments (\"Non Scheduled Encounters\"). Its entry is in _Bold_ below when _Non Scheduled Encounters_ are in the majority. This proportion reflects how much a site doesn't schedule activity.
  * _No Enc Appts_ shows appointments that lack encounters. The _A/NS/C%_ breakdown shows how many of these \"orphan appointments\" are Active, NoShow and Cancelled. In a VistA running VistA Scheduling (_VSE_), cancelled appointments account for the vast majority of the orphans with noshows making up the majority of the balance. Active orphans may reflect a failure of a system synchronization or cleanup task.
  * _Pts_ and _Locns_ show the total number of unique patients and locations involved
        
"""
        gotInP = True if sum(1 for instit in summByInstit if summByInstit[instit]["noIns"]) else False 
        cols = [":Institution", "Encounters"]
        if gotInP:
            noteMU = ""
        else:
            noteMU = "__Note__: no institution in this VistA has in-patient encounters.\n\n" 
        cols.extend(["Sched (AU/M/O)%", "Non Sched %", "No Enc Appts (A/NS/C%)"])
        cols.extend(["Pts", "Locns"])
        tbl = MarkdownTable(cols)
        for instit in sorted(summByInstit, key=lambda x: summByInstit[x]["total"], reverse=True):
            
            institSumm = summByInstit[instit]
            schedAuto = institSumm["schedAutoActive"] + institSumm["schedAutoNoShow"] + institSumm["schedAutoCancel"]
            schedManual = institSumm["schedManualActive"] + institSumm["schedManualNoShow"] + institSumm["schedManualCancel"]
            schedMU = percentBreakdownMU([schedAuto, schedManual, institSumm["schedOnly"]], institSumm["total"])
            nonSchedMU = percentBreakdownMU([institSumm["pceOnly"]], institSumm["total"])
            
            patients = institSumm["patients"].union(summByInstit[instit]["patients"])
            locns = institSumm["locns"].union(summByInstit[instit]["locns"])
            if instit in unmatchedApptsSummByInstit:
                unmatchedApptsSummI = unmatchedApptsSummByInstit[instit]
                unmatchedApptsBDMU = percentBreakdownMU([unmatchedApptsSummI["active"], unmatchedApptsSummI["noShow"], unmatchedApptsSummI["cancel"]], unmatchedApptsSummI["total"], bold=False)
                patients |= unmatchedApptsSummI["patients"]
                locns |= unmatchedApptsSummI["locns"]
                unmatchedApptsMU = "{:,} ({}%)".format(unmatchedApptsSummI["total"], unmatchedApptsBDMU)
            else:
                unmatchedApptsMU = ""
            
            row = [
                "__[{}](#{})__".format(instit, re.sub(r'[ \(\)\.]', '_', instit.lower())),
                reportAbsAndPercent(institSumm["total"], cross["total"])
            ]
            row.extend([
                schedMU,
                nonSchedMU,
                unmatchedApptsMU
            ])
            row.extend([
                reportAbsAndPercent(len(patients), len(allPatients)),   
                reportAbsAndPercent(len(locns), len(allLocns))
            ])
            tbl.addRow(row)
        mu += noteMU + tbl.md() + "\n\n"
        return mu
                           
    def muByInstitByDSSnCreateOption(index, instit, institSumm, dssSummsOfInstit, createOptionSummsOfInstit, printOnly=True):  
        
        imu = "## {}. {} {{#{}}}\n\n".format(index, instit, re.sub(r'[ \(\)\.]', '_', instit.lower()))
        
        # TODO: encounters with cancelled appointments are 'strange', the other
        # side of 'active' appointments with no encounters. These "mis-alignments"
        # need further investigation.
        imu += "This Institution has <span class='countHigh'>{:,}</span> encounters assigned to <span class='countHigh'>{:,}</span> work types (\"stop codes\") for <span class='countHigh'>{:,}</span> patients at <span class='countHigh'>{:,}</span> \"locations\". Of these encounters, <span class='countHigh'>{}</span> are automatically created from appointments (\"Sched Auto\"), <span class='countHigh'>{}</span> are manually created from appointments (\"Sched Manual\"), <span class='countHigh'>{}</span> are marked as scheduled but lack a matching appointment (\"Sched Only\"), while the balance, <span class='countHigh'>{}</span>, are not scheduled (\"Non Sched\"). Both types of scheduled encounter are further broken down into Active (A), NoShow (NS) and Cancelled (C) with the vast majority being active.\n\n".format(
                institSumm["total"],
                len(dssSummsOfInstit),
                len(institSumm["patients"]),     
                len(institSumm["locns"]),
                   
                reportAbsAndPercent(institSumm["schedAuto"], institSumm["total"]),
                reportAbsAndPercent(institSumm["schedManual"], institSumm["total"]),
                reportAbsAndPercent(institSumm["schedOnly"], institSumm["total"]),
                reportAbsAndPercent(institSumm["pceOnly"], institSumm["total"])
            )  
                
        # ################## Do a summary by create option #####################
        
        # TODO: add in some marker on Roll and Scroll or GUI (or not create option => machine user?)
        cotbl = MarkdownTable([":Created By", "Count", "Sched Auto (A/NS/C)%", "Sched Manual (A/NS/C)%", "Sched Only %", "Non Sched %"])
        for createOptionSumm in sorted(createOptionSummsOfInstit, key=lambda x: x["total"], reverse=True):
            # ie/ automatic, manual, orphan (sched only); not scheduled
            totalSched = createOptionSumm["schedAuto"] + createOptionSumm["schedManual"] + createOptionSumm["schedOnly"]
            schedAutoMU = percentBreakdownMU([createOptionSumm["schedAutoActive"], createOptionSumm["schedAutoNoShow"], createOptionSumm["schedAutoCancel"]], createOptionSumm["total"]) 
            schedManualMU = percentBreakdownMU([createOptionSumm["schedManualActive"], createOptionSumm["schedManualNoShow"], createOptionSumm["schedManualCancel"]], createOptionSumm["total"]) 
            schedOnlyMU = percentBreakdownMU([createOptionSumm["schedOnly"]], createOptionSumm["total"]) 
            nonSchedMU = percentBreakdownMU([createOptionSumm["pceOnly"]], createOptionSumm["total"])
            label = createOptionSumm["id"].split(" [")[0]
            if float(totalSched)/float(createOptionSumm["total"]) > 0.5:
                label = "__{}__".format(label)
            cotbl.addRow([
                label,
                createOptionSumm["total"],
                schedAutoMU,
                schedManualMU,
                schedOnlyMU, 
                nonSchedMU
            ])
        imu += "<span class='countHigh'>{:,}</span> options (application types) created encounters for this institution. Options names are __in bold__ where most of the encounters they create are scheduled.\n\n".format(len(createOptionSummsOfInstit))
        imu += cotbl.md() + "\n\n"
    
        # ################# Break stop code down/dominant based grouping ######
        dssSummsByDominantSC = defaultdict(list)

        for dssSumm in dssSummsOfInstit:
                        
            dominantSC = sorted(dssSumm["byServiceCategory"], key=lambda x: dssSumm["byServiceCategory"][x], reverse=True)[0]

            if dominantSC in ["A:AMBULATORY", "X:ANCILLARY PACKAGE DAILY DATA", "T:TELECOMMUNICATIONS"]:
                dssSummsByDominantSC[dominantSC].append(dssSumm)
            else: # include not specified
                dssSummsByDominantSC["O:OTHER"].append(dssSumm)           

        # Allowing no showSC to fit into print 
        def muDSSSummsOfDominantSC(dominantSC, dssSumms, printOnly=True):
        
            cols = [":Stop Code", "Total"]
            if sum(dssSumm["noIns"] for dssSumm in dssSumms):
                cols.append("In Patient")
                showIP = True
            else:
                showIP = False
            cols.extend(["Sched Auto (A/NS/C)%", "Sched Manual (A/NS/C)%", "Sched Only %", "Non Sched %", "Pts", "Locns"]) # Creators removed as too wide
            
            scs = set(sc for dssSumm in dssSumms for sc in dssSumm["byServiceCategory"])
            if printOnly == False:
                cols.append("Options")
                showCreateOption = True
                if len(scs) > 1:
                    cols.append(":Service Categories")
                    showSC = True
                else:
                    showSC = False
            else:
                showCreateOption = False
                showSC = False
                
            tbl = MarkdownTable(cols)
    
            # TODO: pie it with instit overall
            
            for dssSumm in sorted(dssSumms, key=lambda x: x["total"], reverse=True):
                
                totalSched = dssSumm["schedAuto"] + dssSumm["schedManual"] + dssSumm["schedOnly"]
                schedAutoMU = percentBreakdownMU([dssSumm["schedAutoActive"], dssSumm["schedAutoNoShow"], dssSumm["schedAutoCancel"]], dssSumm["total"]) 
                schedManualMU = percentBreakdownMU([dssSumm["schedManualActive"], dssSumm["schedManualNoShow"], dssSumm["schedManualCancel"]], dssSumm["total"]) 
                schedOnlyMU = percentBreakdownMU([dssSumm["schedOnly"]], dssSumm["total"]) 
                nonSchedMU = percentBreakdownMU([dssSumm["pceOnly"]], dssSumm["total"])       
                    
                label = re.sub(r'\_', '/ ', dssSumm["id"].split(" [")[0])
                if float(totalSched)/float(dssSumm["total"]) > 0.5:
                    label = "__{}__".format(label)
                    
                row = [
                    label,
                    dssSumm["total"]
                ]
                if showIP:
                    row.append(dssSumm["noIns"] if dssSumm["noIns"] > 0 else "")
                row.extend([
                    schedAutoMU,
                    schedManualMU,
                    schedOnlyMU,
                    nonSchedMU,
                    
                    len(dssSumm["patients"]),
                    len(dssSumm["locns"])
                    # len(dssSumm["createdBys"]) as too wide
                ])
                if showCreateOption:
                    row.append(muBVC(dssSumm["byOptionCreateUsed"]))
                if showSC:
                    row.append(muBVC(dssSumm["byServiceCategory"]))
                
                tbl.addRow(row)  
                  
            # TODO: add a statement on scheduled/non scheduled for both OR A PIE CHART?
            if len(scs) == 1:
                tmu = "For <span class='countHigh'>{:,}</span> stop codes, there is only __{}__ activity. The names of stop codes with a majority of scheduled encounters are in __bold__ ...\n\n".format(
                    len(dssSumms),
                    dominantSC.split(":")[1]
                )
            elif dominantSC == "O:OTHER":
                tmu = "For <span class='countHigh'>{:,}</span> stop codes, other types of activity dominate. The names of stop codes with a majority of scheduled encounters are in __bold__ ...\n\n".format(
                    len(dssSumms)
                )
            else:
                tmu = "For <span class='countHigh'>{:,}</span> stop codes, __{}__ activity dominates. The names of stop codes with a majority of scheduled encounters are in __bold__ ...\n\n".format(
                    len(dssSumms),
                    dominantSC.split(":")[1]                    
                )            

            tmu += tbl.md() + "\n\n"
                
            return tmu
            
        for dominantSC in sorted(dssSummsByDominantSC, key=lambda x: len(dssSummsByDominantSC[x]), reverse=True):
            imu += muDSSSummsOfDominantSC(dominantSC, dssSummsByDominantSC[dominantSC], printOnly=printOnly)                                    
            
        return """<section>
        
{}</section>

""".format(imu) 
    
    # Encounters means 'direct' ie/ not child or EH encounters. TODO more on them.
    mu = """
    
A copy of VistA __{} [{}]__ was cut on _{}_. For its final 90 full days, starting on _{}_, the system had <span class='countHigh'>{:,}</span> appointments in <span class='countHigh'>{:,}</span> time slots and <span class='countHigh'>{:,}</span> encounters. These involved <span class='countHigh'>{:,}</span> patients at <span class='countHigh'>{:,}</span> institutions with <span class='countHigh'>{:,}</span> \"locations\".

""".format(
        vistaName,
        stationNo,
        cutDate,

        onAndAfterDay,
        totalMatchedAppts + len(unmatchedAppts),
        totalSlots,
        cross["total"], # of encounters
        
        len(allPatients), # cross appt and encounters
        len(summByInstit),
        len(allLocns), # cross appt and encounters
    )
    
    mu += muByInstit(cross, summByInstit, unmatchedApptsSumm, unmatchedApptsSummByInstit)
        
    for i, instit in enumerate(sorted(summByInstit, key=lambda x: summByInstit[x]["total"], reverse=True), 1):
        mu += muByInstitByDSSnCreateOption(i, instit, summByInstit[instit], dssSummsByInstit[instit], createOptionSummsByInstit[instit], printOnly=printOnly)
                
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    print("Serializing Report to {}".format(userSiteDir))
    if not os.path.isdir(userSiteDir):
        raise Exception("Expect User Site to already exist with its basic contents")
    title = "Encounters {}".format(stationNo)
    mu = TOP_MD_TEMPL.format(title, "Encounters") + mu
    open(userSiteDir + "encounters.md", "w").write(mu)
    
# ################################# Reductions/Utils #########################

"""
Overall: want match of apptN, apptL, visit (as encounter) with embedded v provider.
Must include option used to create the encounter so can match that to auto or
manual creation (some created manually after seeing a scheduled appt?). 

TODO: evolve to smaller indexes so can go from appointment perspective indexing
encounter component by apptN or apptL.
"""

"""
Indexing by day of visits, appts (legacy and VSE) and V providers for an 
Encounter Period (onAndAfter, upTo). This enables linking.

Flattened and extraneous properties removed (ie/ work off pter refs/LABEL [ID] vs
full dicts.

TODO:
- credit stop qual -- put in and enforce form (is it added afterwards?)
- child visit date spread (may analyze above)
- BIG: consult tie in or other data_source ref ... 
- make sure enough data for 668, 757 alignments ie/ all leg, all vpe aligned in 668
including WHEN/WHY VISIT or APPT are deleted.
- cover all of https://github.com/vistadataproject/workload/blob/master/prodcloneAnalyze/createWorkDetailMVDM.py includes 409_73 (ie/ counted) and pull of V CPT and V POV
- add Patient Appt
"""
def gatherEncounterDataByDay(stationNo, onAndAfterDay, upToDay):

    jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "legNvpeApptsNVisits.json"
    try:
        gatheredByDay = json.load(open(jsnFl))
    except:
        print("Appt source data not yet gathered by day - organizing relevant 44_003, 409_84, 9000010_06 and 9000010 ...")
    else:
        return gatheredByDay

    # Used for filtering out - same as typer
    def getResourceDay(resource, createProp, onAndAfterDay, upToDay):
        if createProp not in resource:
            return ""
        dtValue = resource[createProp]["value"] if "value" in resource[createProp] else resource[createProp]["label"]
        dayValue = dtValue.split("T")[0]
        if dayValue < onAndAfterDay:
            return ""
        if dayValue >= upToDay:
            return ""
        return dayValue

    """
    9000010: 'date_visit_created' used to decide how far
    back to go (1 year back from onAndAfter) in walk. Unlike appointments, 
    there are so many 9000010's that won't just walk them all to pick
    out those with the matchable appointment/visit time ("visit_admit_datetime")
    in the onAndAfter and until range.
    """
    def gatherVisitsByDay(stationNo, onAndAfterDay, upToDay):
        jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "byDay9000010_{}_{}".format(onAndAfterDay, upToDay)
        try:
            visitsByDay = json.load(open(jsnFl))
        except:
            print("9000010 not yet gathered by day for {} to {} ...".format(onAndAfterDay, upToDay))
        else:
            return visitsByDay

        visitStartDate = (datetime.strptime(onAndAfterDay, "%Y-%m-%d") - relativedelta(years=1)).strftime("%Y-%m-%d") # one year back
        print("Organizing 9000010 starting at {}, one year back from onAndAfter {} ...".format(visitStartDate, onAndAfterDay))
        visitsByDay = defaultdict(list)
        visitsByIEN = {} # for provider and parent linking
        parentIENByChildIEN = {} # for checking below
        unattachedChildVisits = 0
        cntEncountersInDateRange = 0
        cntQualCreditStopVisitsSkipped = 0
        cntTopLevelEncounters = 0
        cntByPackage = Counter()
        createProp = "visit_admit_datetime"   
        startProp = "date_visit_created"
        dataLocn = DATA_LOCN_TEMPL.format(stationNo)     
        store = FMQLReplyStore(dataLocn)
        # Go back one year from onAndAfter which should get all visits created for encounters
        # falling in the onAndAfter/upTo window.
        startAtReply = store.firstReplyFileOnOrAfterCreateDay("9000010", startProp, visitStartDate) 
        if startAtReply == "":
            raise Exception("Can't find 9000010 start reply")
        resourceIter = FilteredResultIterator(dataLocn, "9000010", startAtReply=startAtReply)
        print("... starting at reply {}".format(startAtReply))
        for i, resource in enumerate(resourceIter, 1):
            if i == 1:
                print("\tFirst resource seen {}".format(
                    resource[startProp]["value"] if startProp in resource else "?"
                ))
            if i % 50000 == 0:
                print("\tAnother 50K to {:,}, {:,} taken, {:,} credit stop visit's skipped, last resource day {} - 50K resource create date {}".format(i, cntEncountersInDateRange, cntQualCreditStopVisitsSkipped, resourceDay, resource[startProp]["value"] if startProp in resource else "?"))
            # issue in one of 648
            if "service_category" not in resource:
                print("** Skipping Visit as no setting for 'service_category'")
                continue
            # Keeps it in the window of onAndAfterDay and upToDay
            resourceDay = getResourceDay(resource, createProp, onAndAfterDay, upToDay)
            if resourceDay == "":
                continue # as can't rely on create/ien order == createProp order
            # CREDIT STOP -- qual [SCHED] child
            if "package" in resource and re.match(r'SCHEDULING', resource["package"]["label"]) and "encounter_type" in resource and resource["encounter_type"] == "C:CREDIT STOP":
                cntQualCreditStopVisitsSkipped += 1
                continue
            cntEncountersInDateRange += 1
            resource["_ien"] = resource["_id"].split("-")[1]
            for prop in list(resource):
                if prop in ["_id", "type", "label"]:
                    del resource[prop]
                    continue
                if re.search(r'(flag|cancer|connected|sexual|veteran|112|asia|exposure)', prop):
                    del resource[prop]
                    continue
            flattenPropValues(resource)
            visitsByIEN[resource["_ien"]] = resource
            if "parent_visit_link" in resource:
                parentIEN = re.search(r' \[9000010\-(\d+)', resource["parent_visit_link"]).group(1)
                if parentIEN in visitsByIEN:
                    pvisit = visitsByIEN[parentIEN]
                    if "_child_visits" not in pvisit:
                        pvisit["_child_visits"] = []
                    pvisit["_child_visits"].append(resource)
                    parentIENByChildIEN[resource["_ien"]] = parentIEN
                else:
                    unattachedChildVisits += 1 # should only occur at year start
                continue
            cntTopLevelEncounters += 1
            visitsByDay[resourceDay].append(resource)
            
        print("... finished taking {:,} encounters in date range ({}-{}), {:,} are top level, {:,} children, {:,} in range child visits unattached and dropped".format(
            cntEncountersInDateRange, 
            visitStartDate,
            upToDay,
            cntTopLevelEncounters,
            len(parentIENByChildIEN),
            unattachedChildVisits
        ))
    
        json.dump(visitsByDay, open(jsnFl, "w"))
    
        return visitsByDay
    
    visitsByDay = gatherVisitsByDay(stationNo, onAndAfterDay, upToDay)
        
    """    
    2. V PROVIDER - make subordinate to their referenced visits. Will only do V PROVIDERs
    for visits with admit time in the onAndAfter/upTo range. This can also QA
    if the go-back-a-year visit walk above catches all expected. 

    def gatherVProviderByVisitIEN(stationNo, onAndAfterDay, upToDay, topVisitIENs, childVisitIENs):
    
        print("Organizing 9000010_06 under visits starting at {}, the on and after date ...".format(onAndAfterDay))

    """
        
    def gatherOldApptsByDay(stationNo, onAndAfterDay, upToDay):
    
        jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "byDay44_003_{}_{}".format(onAndAfterDay, upToDay)
        try:
            oldApptsByDay = json.load(open(jsnFl))
        except:
            print("44_003 not yet gathered by day for {} to {} ...".format(onAndAfterDay, upToDay))
        else:
            return oldApptsByDay

        print("Organizing 44_003 (walking all) ...")
        oldApptsByDay = defaultdict(list)
        createProp = "appointment_date_time"
        taken = 0
        for i, resource in enumerate(filteredResultIterator(stationNo, "44_003", useRF=True), 1):
            if i % 50000 == 0:
                print("\tAnother 50K to {:,}, {:,} taken".format(i, taken))
            resourceDay = getResourceDay(resource, createProp, onAndAfterDay, upToDay)
            # filter to window
            if resourceDay == "":
                continue
            taken += 1
            resource["_ien"] = resource["_id"].split("-")[1]
            for prop in ["_id", "patient_container_", "label", "type"]:
                if prop in resource:
                    del resource[prop]
            flattenPropValues(resource)
            oldApptsByDay[resourceDay].append(resource)
        print("... finished processing {:,} 44_003's, {:,} taken".format(i, taken))

        json.dump(oldApptsByDay, open(jsnFl, "w"))
    
        return oldApptsByDay
    
    oldApptsByDay = gatherOldApptsByDay(stationNo, onAndAfterDay, upToDay)

    def gatherNewApptsByDay(stationNo, onAndAfterDay, upToDay):
    
        jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "byDay409_84_{}_{}".format(onAndAfterDay, upToDay)
        try:
            newApptsByDay = json.load(open(jsnFl))
        except:
            print("409_84 not yet gathered by day for {} to {} ...".format(onAndAfterDay, upToDay))
        else:
            return newApptsByDay
        
        print("Organizing 409_84 (walking all, filtering to window) ...")
        newApptsByDay = defaultdict(list)
        createProp = "starttime"
        taken = 0
        for i, resource in enumerate(filteredResultIterator(stationNo, "409_84"), 1):
            if i % 50000 == 0:
                print("\tAnother 50K to {:,}, {:,} taken".format(i, taken))
            resourceDay = getResourceDay(resource, createProp, onAndAfterDay, upToDay)
            # filter to window
            if resourceDay == "":
                continue
            taken += 1
            resource["_ien"] = resource["_id"].split("-")[1]
            for prop in ["_id", "label", "type"]:
                if prop in resource:
                    del resource[prop]
            flattenPropValues(resource) 
            newApptsByDay[resourceDay].append(resource)
        print("... finished processing {:,} 409_84, {:,} taken".format(i, taken))
        
        json.dump(newApptsByDay, open(jsnFl, "w"))

        return newApptsByDay

    newApptsByDay = gatherNewApptsByDay(stationNo, onAndAfterDay, upToDay)
                    
    print("Returning Organized Data")     
    return visitsByDay, oldApptsByDay, newApptsByDay
        
"""
Post gathering by day, hone down matches by location/patient too
"""               
def indexEncounterDataByDateLocationPatient(stationNo, visitsByDay, oldApptsByDay, newApptsByDay):

    # ######### Organize Appts + Encounters into Time/Locn/Patient Hierarchy #######
    # ... TODO: may move this into the Reducer ie/ make the reducer do this
    
    # A/C for older _ sub'ed for / in locn and if locnRef is delegate with different
    # id but same label ie/ reduce to a / using label
    def toLocnLabel(locnRef):
        return re.sub(r'\_', '/', locnRef.split(" [")[0])
    
    matchesDayTimeLocationPatient = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(list)))))
    # from union of old and new appts (not visit only days)
    days = set(oldApptsByDay).union(set(newApptsByDay))
    for day in days:
        if day not in oldApptsByDay:
            continue
        for res in oldApptsByDay[day]:
            if not ("appointment_container_" in res and "appointment_date_time" in res and "patient" in res):
                print("** Warning - bad old appt missing expected props")
                continue
            tm = res["appointment_date_time"].split("T")[1]
            locnLabel = toLocnLabel(res["appointment_container_"])
            patient = res["patient"]
            if "appointment_cancelled" in res:
                typ = "apptLC"
            else:
                typ = "apptL"
            matchesDayTimeLocationPatient[day][tm][
            locnLabel][patient][typ].append(res)  
    for day in days:
        if day not in newApptsByDay:
            continue
        for res in newApptsByDay[day]:
            if not ("resource" in res and "starttime" in res and "patient" in res):
                print("** Warning - bad new appt missing expected props, resource | starttime | patient")
                continue
            tm = res["starttime"].split("T")[1]
            locnLabel = toLocnLabel(res["resource"])
            patient = re.sub(r'9000001', '2', res["patient"])
            if "cancel_datetime" in res:
                typ = "apptNCancelled"
            elif "noshow" in res:
                typ = "apptNNoShow"
            else:
                typ = "apptN"
            matchesDayTimeLocationPatient[day][tm][
            locnLabel][patient][typ].append(res)
    # REM: reduction doesn't have credit stop only ([SCHED]/parent) visits
    # ... TOP SCHED, TOP PCE, child PCE (under either top) as never have a child
    # SCHED 'encounter'
    # CHANGE COMING: may exclude EH visits from RED
    totalMatchableTopEncounters = 0
    totalMatchableEncounters = 0
    noTimeVisitsByPackage = Counter() # + double check typer that pkg mand in 757 too
    noHospitalLocationVisitsByPackage = Counter()
    excludedVisitsAsMissingProps = 0
    excludedScheduledVisitsAsMissingProps = 0
    countExcludedVisitsAsMissingPropsSC = Counter()
    totalVisitsEH = 0
    totalChildren = 0
    totalNoTimeTops = 0
    for day in days:
        if day not in visitsByDay:
            continue
        for res in visitsByDay[day]:
            if not ("hospital_location" in res and "visit_admit_datetime" in res and "patient_name" in res):
                if res["service_category"] != "E:EVENT (HISTORICAL)":
                    raise Exception("Expect key props in all PCE and SCHED visits - only EH excluded for missing props")
                continue
            # Can exclude as key props will all agree with parent
            if "parent_visit_link" in res:
                if re.match(r'SCHEDULING', res["package"]): # extra - no SCHED is a child
                    raise Exception("Scheduling with parent!")
                totalChildren += 1
                continue # let's just have top
            # note: visit_admit_date_time must support tm
            if not re.search(r'T', res["visit_admit_datetime"]):
                noTimeVisitsByPackage[res["package"]] += 1
                totalNoTimeTops += 1
                continue
            tm = res["visit_admit_datetime"].split("T")[1]
            if "hospital_location" not in res:
                noHospitalLocationVisitsByPackage[res["package"]] += 1
                continue
            locnLabel = toLocnLabel(res["hospital_location"])
            # TODO: BUT perhaps exclude res's w/o it (saw in 648)
            if res["service_category"] == "E:EVENT (HISTORICAL)": # some have all keys props so exclude em all
                if "package" in res and re.match(r'SCHEDULING', res["package"]):
                    raise Exception("Dont expect EVENT HISTORICAL service_category with package SCHEDULING")
                totalVisitsEH += 1
                # doing as EH Visit to avoid 
                matchesDayTimeLocationPatient[day][tm][
            locnLabel][patient]["ehVisit"].append(res) 
                continue
            # Exclude no locn, no admit
            totalMatchableEncounters += 1
            patient = re.sub(r'9000001', '2', res["patient_name"])
            if "package" in res and re.match(r'SCHEDULING', res["package"]):
                matchesDayTimeLocationPatient[day][tm][
            locnLabel][patient]["visit"].append(res) 
                continue
            # ie/ non schedule appts - inc with no package!
            matchesDayTimeLocationPatient[day][tm][
            locnLabel][patient]["visitNS"].append(res)
    print("Summary of Visits: total matchable encounters {:,} - {:,} children, {:,} top EHs and {:,} no time tops are not in matcheable categories".format(totalMatchableEncounters, totalChildren, totalVisitsEH, totalNoTimeTops))
    if len(noTimeVisitsByPackage):
        print("** WARNING: no time in visit_admit_time by pkg {}".format(noTimeVisitsByPackage))
    if len(noHospitalLocationVisitsByPackage):
        print("** WARNING: no location by pkg: {}".format(noHospitalLocationVisitsByPackage))
        
    return matchesDayTimeLocationPatient
    
def categorizeVisitAppts(info, noneApptOnlyDetails=None, matchedDetails=None, matchedDetails2ByCatag=None, visitCreateBorAStart=None):

    # Reduce to singular of a type if can        
    def oneAppt(info, apptPrefix="apptN"):
        apptProps = [prop for prop in info if re.match(apptPrefix, prop)]
        if len(apptProps) == 0:
            return None
        # Strict for apptN - looser (need apptL) for apptL (TODO: tighten)
        if len(apptProps) > 1:
            if apptPrefix == "apptN" or apptPrefix not in info:  
                # USED TO BE AN EXCEPTION BUT ISSUE FOR POR 663 TODO
                print("*** Ambiguous multiple {} matches".format(apptPrefix, json.dumps(apptProps)))
            # apptProp = apptPrefix
            apptProp = apptPrefix if apptPrefix in apptProps else apptProps[0]
        else:
            apptProp = apptProps[0]
        # Strict for apptN - looser - pick bigger for apptL (TODO: tighten)
        if len(info[apptProp]) > 1:
            if apptPrefix == "apptN": # Warning for apptN (weakened exp for 663)
                print("** Ambiguous multiple {}s at same time".format(apptPrefix))
            ressBySize = sorted(info[apptProp], key=lambda x: len(x), reverse=True)
            appt = ressBySize[0] # ie/ with checkout etc     
        else:
            appt = info[apptProp][0]
        return appt
        
    """
    Matching appt and visit DT props to get nearest time when visit created. Is
    it created when appt is or at checkin or at checkout or ?
    
    EXTRA: Events would create, checkout, cancel etc. ie/ a state table.
    ... fill in this and other pattern notes
        apptL:
        - date_appointment_made == visit:date_visit_created
        - appointment_date_time = visit:visit_admit_datetime
        - data_entry_clerk == visit:created_by_user
        - checked_out=apptN:checkout
        apptN:
        - starttime = visit:visit_admit_datetime (and apptL)
        and maybe visit:date_last_modified == checked
    """
    def matchCreationDateTimes(apptN, apptL, visit):

        # Should be day after 0:50
        # ? what option used in appt create here? ie/ what app? Why all CC 
        # like this?
        if re.match(r'SDAM BACKGROUND', visit["option_used_to_create"]):
            return None

        # Issue that no match time as vista can be out by seconds - when
        # out indicates code sequential using NOW as opposed to just cloning (apptN-SCHED)
        def squeezeTimes(dtTimeCollection):
            dtMap = {}
            dts = [dt for dt in dtTimeCollection if re.search(r'\dZ', dt)]
            # within 60 seconds => take lower
            mappedDTs = {}
            cnt = 0
            for i, dt in enumerate(sorted(dts), 1):
                try:
                    dtDT = datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S")
                except:
                    print("** PROBLEM BAD DATE TIME: {}".format(dt))
                    continue
                cnt += 1
                if cnt == 1:
                    prevDT = dtDT
                    prevdt = dt
                    continue
                if dtDT - prevDT < timedelta(seconds=60):
                    mappedDTs[dt] = prevdt
                    dtTimeCollection[prevdt].extend(dtTimeCollection[dt])
                    del dtTimeCollection[dt]
                    continue
                prevDT = dtDT
                prevdt = dt
            return dtTimeCollection 
           
        # For two and three 
        def isstr(s):
            try:
                return isinstance(s, basestring)
            except NameError:
                return isinstance(s, str)

        appt = apptN if apptN else apptL
        typedDTPropsByDT = defaultdict(list)
        for prop in appt:
            # know not relevant
            if prop == "desired_date_of_appointment":
                continue 
            # walkin can be bool
            if not isstr(appt[prop]):
                continue
            if re.search(r'\dZ$', appt[prop]):
                typedDTPropsByDT[appt[prop]].append("appt:" + prop)
                continue
            if re.match(r'\d\d\d\d\-\d\d\-\d\d$', appt[prop]):
                typedDTPropsByDT[appt[prop]].append("appt:" + prop + "_day")
        visitDtPropsByDT = defaultdict(list)
        for prop in visit:
            if not isstr(visit[prop]):
                continue
            if re.search(r'\dZ$', visit[prop]):
                typedDTPropsByDT[visit[prop]].append("visit:" + prop)
                if prop == "date_visit_created": 
                    dayOnlyDT = visit[prop].split("T")[0]
                    typedDTPropsByDT[dayOnlyDT].append("visit:" + prop + "_day")
                continue
            if re.match(r'\d\d\d\d\-\d\d\-\d\d$', visit[prop]):
                typedDTPropsByDT[visit[prop]].append("visit:" + prop + "_day")
        typedDTPropsByDT = squeezeTimes(typedDTPropsByDT)
        # Stick to matchable props
        mtch = None
        for dt in typedDTPropsByDT:
            apptProps = [prop.split(":")[1] for prop in typedDTPropsByDT[dt] if re.match(r'appt:', prop) if re.search(r'(\_made|check)', prop.split(":")[1])]
            if len(apptProps) == 0:
                continue
            if "visit:date_visit_created" in typedDTPropsByDT[dt]:
                mtch = ("date_visit_created", "/".join(sorted(apptProps)))
            elif "visit:date_visit_created_day" in typedDTPropsByDT[dt]:
                mtch = ("date_visit_created_day", "/".join(sorted(apptProps)))
        return mtch
        
    # TODO: just make this and logic at end func be the func ie/ no need
    # for explicit enum of SchedVisit and its options etc. That was just for
    # dev
    def categorizeMatch(visitType, visit, apptType, apptN, apptL):
        if "option_used_to_create" not in visit: 
            catag = "{}_NOOPTION".format(visitType)  
            matchedDetails["apptTypeByCatag"][catag][apptType] += 1 
            return catag
        optionUsedToCreateLabel = visit["option_used_to_create"].split(" [")[0]
        catag  = "{}_{}".format(visitType, optionUsedToCreateLabel)     

        matchedDetails["apptTypeByCatag"][catag][apptType] += 1

        if matchedDetails2ByCatag != None:
            mtch = matchCreationDateTimes(apptN, apptL, visit)
            if mtch == None:
                apptVisitCreationMatch = "NONE"
            else:
                apptVisitCreationMatch = mtch[1]
            matchedDetails2ByCatag[catag][apptVisitCreationMatch] += 1
            
        return catag
        
    # Matching is from Visit back to Appts
    def matchScheduledVisit(info, apptType, apptN, apptL):
            
        if len(info["visit"]) != 1:
            """
            Saw where second visit as if first appt at exact time was deleted
            and the older visit created lingered. 
            """
            return "SCHED_AMBIG_VISIT"
                
        visit = info["visit"][0] 
                                
        """
        TODO: O:OCCASION OF SERVICE for MRI etc and for that service_category is
        X:ANCILL. O:OCC's are NON COUNT
        """
        if not ("encounter_type" in visit and visit["encounter_type"] in ["P:PRIMARY", "O:OCCASION OF SERVICE"]):
            raise Exception("Expected SCHED Visit match to have encounter type P:PRIMARY or O:OCCASION OF SERVICE")
                                            
        if "option_used_to_create" not in visit: 
            catag = "SCHED_NOOPTION"  
            matchedDetails["apptTypeByCatag"][catag][apptType] += 1 
            return catag
                                
        optionUsedToCreateLabel = visit["option_used_to_create"].split(" [")[0]
        catag  = "SCHED_{}".format(optionUsedToCreateLabel)     
                
        # Know there is an APPTN
        matchedDetails["apptTypeByCatag"][catag][apptType] += 1

        if matchedDetails2ByCatag != None:
            mtch = matchCreationDateTimes(apptN, apptL, visit)
            if mtch == None:
                apptVisitCreationMatch = "NONE"
            else:
                apptVisitCreationMatch = mtch[1]
            matchedDetails2ByCatag[catag][apptVisitCreationMatch] += 1
            
        """
        KEY: NOT AS TELLING AS MAYBE CREATE APPT AFTER VISIT TIME TOO
        .. ie/ hence SDECRPC which aligns APPT and VISIT will have 'surprising'
        AFTER
        """
        if visitCreateBorAStart != None:
            visitCreate = visit["date_visit_created"] 
            visitStart = visit["visit_admit_datetime"]
            if not re.search(r'T', visitCreate):
                visitStart = visitStart.split("T")[0]
            ordr = "BEFORE" if visitCreate <= visitStart else "AFTER"
            visitCreateBorAStart[catag][ordr] += 1
                    
        # TODO: can reduce the following to comment of most frequent (ie/ move from
        # above comment with comparison!!! 
        
        # Important -- the following catag breakdown are placeholders to highlight main
        # catags and allow checks in the future -- effectively the code just returns
        # 'catag' now                   
                        
        """
        Pattern TODO: removed for now
        
        data_entry_clerk == created_by_user but apptN code messes up as 
        no separate slots for various other setters.
            
        # Appears that last changer updates 'entry clerk' ie/ original lost
        if not ("checkin" in apptN or "noshow_by_user" in apptN or "cancelled_by_user" in apptN or "walkin" in apptN) and apptN["data_entry_clerk"] != visit["created_by_user"]:
            raise Exception("Expected SCHEDULED creator to match in appt and visit if no checkin or other mods")        
        """
            
        """
        In general, there is apptN-apptL-visit matching when using SDECRPC
        (SPO 65986 of 68307 of SDECRPC do have apptL)
        
        ApptL there or not
        - NOT: in general (1661 vs 74), apptN cancelled, date before appt start then 
        none (deleted?)
        - THERE: in general (1517 vs 8), apptN noshow, then apptL still there ie/ 
        not removed
        - ... misc others but low # 650+ and could just be quirks -- can't see pattern
        
        TODO Pattern fill in:
        - aligning create times
          - time create apptN is effected by checkin, checkout
          - time apptL less so ie/ still matches visit create time but not all the time
        - why O:OCCASION OF SERVICE (LAB-X etc?) -- plays into CATSCAN etc X:ANCIL, O:OCC
        - consult_links: appt_request_type is to 123, explicit in apptL etc.
        - a/c for misc non apptL others ... lowish # but can't see pattern 
        (based on locn? See typer)
        """
        if re.match(r'SDECRPC', visit["option_used_to_create"]):
                
            return catag
        
        """
        BG: generally no apptL - just apptN; run next day after midnight.
        (SPO only 608 have apptL, 9493 not) 
        
        ? is what created the appointment? Unlike SDECRPC, don't know per se. Presumably
        this a/cs for exceptions on apptL.
        
        TODO: WHY BG needed - due to options NOT making visits? -- is its lack of
        apptLs <=> orphan apptL #'s ie/ different matching? appt date slightly out
        in apptLs in the option create.
        
        KEY CONFIG: only one created_by_user -- setup in option
        """        
        if re.match(r'SDAM BACKGROUND', visit["option_used_to_create"]):

            if not re.search(r'T0[0123]', visit["date_visit_created"]):
                raise Exception("Expected SDAM BG job to run in the three hours after midnight (the next day)")    
                                    
            return catag # doesn't say what made the appt -> encounter as diff time
                
        """
        Roll and Scroll (see users) ala CPRS with Manual Creation? 
                
        See CC visits created with checkout=starttime and 10 minute duration. 
        The appt and visit are created AFTER the appt time => manual process 
        (SUBCATEGORY?)
                
        Some have protocol: 'SDCO APPT CHECK OUT' in visit (ie/ triggered?)
        [an option 101 for 'checking out an appt']
                
        TODO: see CC appts created like this with 10 min duration, entered
        AFTER appt time, source is consult and ApptN only (not ApptL)
        """
        if re.match(r'SDAM APPT MGT', visit["option_used_to_create"]):
                                                                
            return catag
                
        """
        Few, no apptL
        """
        if re.match(r'SDM \[', visit["option_used_to_create"]):
                                                
            return catag
                
        """
        Few, all apptL
        """
        if re.match(r'SDMULTIBOOK', visit["option_used_to_create"]):    
                
            return catag
                
        """
        Mainly INP and dss of ADMIN but more nuance there.
        
        Few ApptL - 668 had none, 663 had DG ADMIT PATIENT
                
        TODO: in hosp care => appt may pre-exist but no checkin/out and match
        is made in an in hospital option earlier in the day of start.
        ... do more when do in hospital care more.
                           
                DG DISCHARGE PATIENT is one: straight apptN (no L) and 1 dep visit w/o
        a checkin/out or other stateful changes
        """
        if re.match(r'DG ', visit["option_used_to_create"]):
        
            if apptL:
                # 663 has DG ADMIT PATIENT but 668 doesn't
                print("** Don't expect SCHED DG ({}) to have ApptL".format(visit["option_used_to_create"]))         
                    
            return catag # actually a set of DG's possibly
                
        """
        PXCE ENCOUNTER ENTRY ...
        ... a series of options. Roll and Scroll. Says allow V related files 
        to be updated.
        
        Tiny - 2 apptL, 1 w/o 
                
        Note: dss probably ADMIN 
        """
        if re.match(r'PXCE ENCOUNTER', visit["option_used_to_create"]):         
                
            return catag
                
        return catag # SHOULD BREAK DOWN
        
    """
    Much less likely - where an app is used to (or roll and scroll) to fill in a
    PCE visit to match an appt
    """
    def matchPCEVisit(info, apptType, apptN, apptL):
                                     
        """
        NON SCHED (PCE) Encounter for Scheduled Appt
        
        Notes:
        - protocol rare (hardly ever) and PXCE SDAM UPDATE ENCOUNTER
        - encounter_type (all as CREDIT STOP out): P:PRIMARY, O:OCCASION, A:ANCILLARY, 
        """
        if not ("visitNS" in info and len(info["visitNS"]) == 1): 
            return "PCE_AMBIG_VISIT"
        visitNS = info["visitNS"][0]
                            
        """
        We see PXCE SDAM ... w/o encounter_type
        
        Interplay with service_category: X:ANCILL with O:OCC and plays into non count
        and MRI, CAT SCAN etc.
        """
        if "encounter_type" in visitNS and visitNS["encounter_type"] not in ["P:PRIMARY", "O:OCCASION OF SERVICE", "A:ANCILLARY"]:
            raise Exception("Expected Non SCHED Visit (NS) match to have encounter type P:PRIMARY or O:OCCASION OF SERVICE or A:ANCILLARY ")    
                    
        if "option_used_to_create" not in visitNS:
            catag = "PCE_NOOPTION"
            matchedDetails["apptTypeByCatag"][catag][apptType] += 1
            return catag  
            
        optionUsedToCreateLabel = visitNS["option_used_to_create"].split(" [")[0]
        catag  = "PCE_{}".format(optionUsedToCreateLabel)     
                
        # Know there is an APPTN
        matchedDetails["apptTypeByCatag"][catag][apptType] += 1 
            
        if matchedDetails2ByCatag != None:
            mtch = matchCreationDateTimes(apptN, apptL, visitNS)
            if mtch == None:
                apptVisitCreationMatch = "NONE"
            else:
                apptVisitCreationMatch = mtch[1]
            matchedDetails2ByCatag[catag][apptVisitCreationMatch] += 1
            
        if visitCreateBorAStart != None:
            visitCreate = visitNS["date_visit_created"] 
            visitStart = visitNS["visit_admit_datetime"]
            if not re.search(r'T', visitCreate):
                visitStart = visitStart.split("T")[0]
            ordr = "BEFORE" if visitCreate <= visitStart else "AFTER"
            visitCreateBorAStart[catag][ordr] += 1
            
        # Important -- the following catag breakdown are placeholders to highlight main
        # catags and allow checks in the future -- effectively the code just returns
        # 'catag' now                                      
        
        """
        Vast Majority of PCE matches
        
        Most lack apptL (11789 vs 1723; of none, 928 cancelled and 186 noshow)
        
        # Add many of the same as DSS
        """
        if re.match(r'OR CPRS GUI CHART', optionUsedToCreateLabel):
                       
            return catag
        
        """
        apptL usually there: 12 has; 12 cancelled => would be deleted; 1 missing
        
        ie/ if check for "cancelled_by_user" in apptN. Also shows that most dental
        don't come this route at all. Perhaps cancelling resets option in visit too?
        """
        if re.match(r'DENTV DSS DRM GUI', optionUsedToCreateLabel):
                                
            # alternative: "noshow_by_user" and note that time then of visit
            # create doesn't seem to match even the noshow recording ie/ seems
            # to be situation where VISIT setup manually by DSS GUI so comments
            # could be made on noshow!
            # ... ie/ not really a match per se. checkout could be the same ie/
            # commenting on a visit that may never have happened even if didn't
            # note a noshow.  
                    
            return catag
                
        # ED TRACKING SYSTEM - got RPCs (calling out as RPCs)
        if re.match(r'EDPF TRACKING SYSTEM', optionUsedToCreateLabel):                
                
            return catag
                
        """
        # EC GUI CONTEXT - Event Capture GUI option
        # ... note that checkout is present
        
        noApptL (11 to 1 ie/ only 1 has it)
        """
        if re.match(r'EC GUI CONTEXT', optionUsedToCreateLabel):
            
            if visitNS["encounter_type"] != "A:ANCILLARY":
                raise Exception("Expected scheduled visit from EC GUI to be Ancillary")           
            
            return catag
                
        """
        Never apptL
        """
        if re.match(r'DVBA CAPRI GUI', optionUsedToCreateLabel):               
            
            return catag
                
        """
        Never apptL
        """
        if re.match(r'DSIU MENTAL HEALTH SUITE', optionUsedToCreateLabel):              
                
            return catag
                
        return catag
                           
    if len(info) == 0:
        return "NONE_EMPTY"
                                      
    """
    Remove child PCEs, EH PCEs and PCEs if no appts
    """    
    def pruneNS(info): 
        # Cleanup visitNS - exclude E:EVENT, children
        if "visitNS" in info:
            visitNSR = [visitNS for visitNS in info["visitNS"] if not re.match(r'E:EVENT', visitNS["service_category"])]
            visitNSR = [visitNS for visitNS in visitNSR if "parent_visit_link" not in visitNS]
            if len(visitNSR):
                info["visitNS"] = visitNSR
            else:
                del info["visitNS"]
        return info
    info = pruneNS(info)
    
    # In case others in info for calcs ... using 'visit' and not ehVisit means EH won't count
    keys = set([key for key in info if re.match(r'(appt|visit)', key)])
    
    if len(keys) == 0:
        return "NONE_NOTMATCHABLE" # include EH only! TODO
    
    if keys == set(["visitNS"]):
        return "NONE_PCEONLY"
        
    if keys == set(["visit"]):
        return "NONE_SCHEDONLY"
        
    """
    No Match as missing either encs or appts to match
    ... may remove as should only work if both appt and enc
    """               
    if sum(1 for k in keys if re.match(r'visit', k)) == 0:
        if noneApptOnlyDetails != None:
            noneApptOnlyDetails["/".join(sorted(info))] += 1
        return "NONE_APPTONLY"
    
    apptTypes = []             
    apptN = oneAppt(info, "apptN")
    if apptN:
        apptTypes.append("apptN")
    apptL = oneAppt(info, "apptL") 
    if apptL:
        apptTypes.append("apptL")
    apptType = "/".join(apptTypes)
    
    if apptType == "" or sum(1 for k in keys if re.match(r'visit', k)) == 0:
        raise Exception("Expect to be matchable at this stage")
                
    if "visit" in info: # visit == visitS
        return matchScheduledVisit(info, apptType, apptN, apptL)
    return matchPCEVisit(info, apptType, apptN, apptL)

# ###################### Per 9000010 Enforce #####################

"""
Note: may move to a general enforcer category/ sub type

Quad of: PCE | SCHED, CHILD | TOP
where Child SCHED under either is credit stop qual and it is NOT an Encounter.

... using Typer red as it should

In order of importance: PCE/TOP, SCHEDULE/TOP and then (ignore qualifying SCHED-CHILD DSS) a small number of real children most of which are E:HIST.

TODO MORE: beyond called for below (reproduce visits from rawer data)
- BIG: DSS report and Locn -- see that visit dss (inc [SCHED] child qual) from locn DSS's
  ... will need to work dates there.
- outside_location and see where it appears (use hosp_locn and in_out breaks)
- If matched with cancelled or noshow appt, is there any form? ie/ would it be an auto-created that can't be deleted?
  ... part of ability to recreate VDM (visit, 409_84 etc) from MVDM of Encounter, Appt

NOTE: move to pure Encounter setup or WL as beyond Scheduled Encounters (SCHED/TOP)
"""
def enforce9000010Quad(stationNo):

    """
    - Basic is package (for x 4) and parent_package_label/#parent_visit_link (combo as can't work out package for some) 
    - hospital_location adds For DSS + Behavior and interplays with patient_status_in_out
    - type__003 allows VA or OTHER and outside_location stuff (bonus) 
    ie/ lot's of options for different combos
    """
    _all, sts = splitTypeDatas(stationNo, "9000010", reductionLabel="2019ToCutE", expectSubTypeProperties=["package", "#parent_visit_link", "parent_package_label", "hospital_location", "type__3", "patient_status_in_out"]) 
    
    """
    'Fake visit' child - just CREDIT STOP, acting as a stop qualifier
    ... this is NOT an encounter and should be excluded during encounter matching
    
    Other properties will be copies of its parent (identical loc_of_encounter, hospital_location, identical type__03, identical eligibility)
    
    Dates not enforced here but checked in manual walk
    - visit_admit_date_time same
    - date_last_modified == date_visit_created if full or == day of visit_created
      ... sometimes only day set for modified
    
    Key: why package set to SCHEDULING. Many seem to be caused by PCE parent/CPRS
    using 
    
    - option_used_to_create
      - for SCHED parent: SCHED options
      - for PCE parent: mainly CPRS
    - service_category: mainly X:ANCILL for both
    
    IMPORTANT: for PCE Parent, most have the optional protocol property set to SDAM PCE EVENT which must be triggers during create of CPRS (and others?) and it must set PACKAGE to SCHEDULING. ie/ fake package setting, almost a bug == quirk.
    """
    def enforceScheduleChild(st, parentPackageLabel):
    
        if singleValue(st, "encounter_type") != "C:CREDIT STOP":
            raise Exception("SCHILD: expect CREDIT STOP only ET")
            
        # TODO: is this 409_68?
        if singleValue(st, "dependent_entry_count") != "1":
            raise Exception("SCHILD: expect only 1 dependent entry")
                    
        # Key of fake setting of SCHEDULING as package - other package calls into a protocol of SCHEDULING
        if "protocol" in st and parentPackageLabel != "SCHEDULING":
            print("** INFO: protocol set for {} Sched Child under {} - {}".format(
                parentPackageLabel,
                reportAbsAndPercent(st["protocol"]["count"], st["_total"]),
                muBVC(st["protocol"]["byValueCount"])
            ))
                    
    """
    'Real' child and relatively simple/few patterns
    - under PCE: in-hospital, telephony and E:HIST (low ANCILL too?)
    - under SCHED: all E:HIST <------- only used for E:HIST under scheduled
        
    Note: very little used as E:HIST usually stands alone (PCE/TOP)
    """
    def enforcePCEChild(st, parentPackageLabel):
    
        if "C:CREDIT STOP" in st["encounter_type"]["byValueCount"]:
            raise Exception("PCHILD: CREDIT STOP not expected in ET")
            
        if not re.match(r'OR CPRS GUI CHART', singleValue(st, "option_used_to_create")):
            raise Exception("PCHILD: only expect option used of CPRS")
            
        # AND only one in 668 when parent is SCHED
        if "E:EVENT (HISTORICAL)" not in st["service_category"]["byValueCount"]:
            print("** WARNING: expected E:HIST service_category to be in PCE Child options")
        
    """
    - option_used_to_create: SCHED options (and DG) but NO HL7 etc 
      ... todo: small rogue CPRS
    - encounter_type: PRIMARY dominates and some OCCASION OF SERVICE (TODO more)
    - service_category: AMB, TELECOM, HOSP stuff (but some X:ANCILL too) ... note no E:HIST
    Note: very low use of 'protocol' so ignoring for now
    Note: very few outside_location but follow up on em
    """
    def enforceScheduleTop(st):

        if "C:CREDIT STOP" in st["encounter_type"]["byValueCount"]:
            raise Exception("S/-: CREDIT STOP not expected in ET")
            
        if "E:EVENT (HISTORICAL)" in st["service_category"]["byValueCount"]:
            raise Exception("S/-: don't expect service_category of E:HIST")
    
        if "outside_location" in st:
            raise Exception("S/-: don't expect outside locations for Scheduled (TOP) encounters")
    
        if set(st["encounter_type"]["byValueCount"]) != set(["P:PRIMARY", "O:OCCASION OF SERVICE"]):
            print("** WARNING: expect only Primary and OOS encounter type for Schedule Top")
                                                
    """
    - option_used_to_create: many but CPRS dominates (and then a big drop to LRVR ... and alot more) ie/ THIS IS WHERE variety kicks in for various apps
    - service_category: E:HIST, A:AMB (but some ANCILL too) ...
    - very low use of protocol so ignoring
    
    ... large variety and would need more breakdown itself but note a lot of E:HIST ie
    presumably many added in.
    
    Note: only one with full outside_location (very few) but follow up
    """
    def enforcePCETop(st):

        if "C:CREDIT STOP" in st["encounter_type"]["byValueCount"]:
            raise Exception("P/-: CREDIT STOP not expected in ET")
                            
    # Important: custom can't map all parent links to a parent for its package =>
    # link is superset and for processing we will exclude st's where there is a link
    # but no label.
    stsPP = combineSubTypes(sts, subTypePropsSubset=["package", "#parent_visit_link", "parent_package_label"])
    typeCounts = Counter()
    for st in stsPP:
        pkg = singleValue(st, "package")
        if "parent_visit_link" in st:
            # ignore st if parent pkg label missing when there is a link
            if "parent_package_label" not in st:
                continue 
            ppkg = singleValue(st, "parent_package_label")
            # Not real visit - lazy qual credit stop
            if re.search(r'SCHEDULING', pkg):
                enforceScheduleChild(st, ppkg)
                typeCounts["ChildSched" + ppkg[0]] = st["_total"]
            else:
                enforcePCEChild(st, ppkg)
                typeCounts["ChildPCE" + ppkg[0]] = st["_total"]
            continue
        if re.search(r'SCHEDULING', pkg):
            enforceScheduleTop(st)
            typeCounts["Sched"] = st["_total"]
            continue
        enforcePCETop(st)
        typeCounts["PCE"] = st["_total"]
        
    totalEncs = typeCounts["PCE"] + typeCounts["Sched"] + typeCounts["ChildPCES"] + typeCounts["ChildPCEP"]
    totalTopEncs = typeCounts["PCE"] + typeCounts["Sched"]
    print("## Summary Totals")
    print("... note doesn't count the visits who parent visit link couldn't be resolved")
    print("Total Encounters (not SCHED CHILD/DSS quals): {:,}".format(totalEncs))
    print("Total Encounters Top: {}".format(
        reportAbsAndPercent(totalTopEncs, totalEncs)
    ))
    print("Children as % of Top PCE: {}".format(
        reportAbsAndPercent(typeCounts["ChildPCEP"], typeCounts["PCE"])
    ))
    print("Children as % of Top Sched: {}".format(
        reportAbsAndPercent(typeCounts["ChildPCES"], typeCounts["Sched"])
    ))
    print("Qualifying Visits as % of Top PCE: {}".format(
        reportAbsAndPercent(typeCounts["ChildSchedP"], typeCounts["PCE"])
    ))
    print("Qualifying Visits as % of Top Sched: {}".format(
        reportAbsAndPercent(typeCounts["ChildSchedS"], typeCounts["Sched"])
    ))
    
# ################################# DRIVER #######################

def main():

    assert sys.version_info >= (3, 6)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO")

    ensureWebReportLocations(stationNo)

    meta = metaOfVistA(stationNo)
    vistaName = "VistA" if "name" not in meta else meta["name"]
    cutDate = meta["cutDate"]
    cutDateDT = datetime.strptime(cutDate.split("T")[0], "%Y-%m-%d")
    day90DateDT = cutDateDT - relativedelta(days=90)
    day90Date = datetime.strftime(day90DateDT, "%Y-%m-%d")
    print("For {}, cutdate {} is upTo and 90 day back, {}, is on or after".format(stationNo, cutDate, day90Date))
    onAndAfterDay = day90Date
    upToDay = cutDate
    
    # Here until whole report moves to typer and E
    allThere, details = checkDataPresent(stationNo, [

        {"fileType": "9000010", "check": "ALL"},
        {"fileType": "9000010_06", "check": "ALL"},
        {"fileType": "409_84", "check": "ALL"},
        {"fileType": "44_003", "check": "ALLRF"}

    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))
        
    # expectations of 9000010 enforced - TODO: change to merge into DataByDay's 9000010 
    # walk
    # enforce9000010Quad(stationNo)

    print("Gather | Load encounter data by day ...")
    visitsByDay, oldApptsByDay, newApptsByDay = gatherEncounterDataByDay(stationNo, onAndAfterDay, upToDay) 
    print("... loaded")
    print("Index encounter data (appts and visits) using location, patient, time ...")
    indexedDayTimeLocationPatient = indexEncounterDataByDateLocationPatient(stationNo, visitsByDay, oldApptsByDay, newApptsByDay)
    print("... indexed")
    
    webReportEncounters(stationNo, vistaName, indexedDayTimeLocationPatient, cutDate, onAndAfterDay, upToDay)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter

"""           
Station Numbers

- 3 digits (say 5+ too but why?)
- appended with a suffix such as 'A' or '9' (9 is a special case, representing a nursing home subdivision).
- 200-series station numbers, which are only allocated to Austin Information Technology
Center (AITC) systems
(from vistalink manual: https://www.va.gov/vdl/documents/Infrastructure/VistALink/vistalink_1_6_dg.pdf)
"""
SSNO_DEFINITIONS = { # see p81 for main SSNs ... list all and role
    "101": "Central Office, Washington, DC",
    "104": "AAC, Finance Center",
    "200": "AAC",
    
    # Go with 100 for CAPRI ie/ pairing - logon first with 100 and then back?
    "318": "Regional Office - Winston-Salem, NC", # see p82 for regional offices and their 3**
    "377": "Regional Office - San Diego, CA",
    
    "436": "FORT HARRISON MT",
    "442": "CHEYENNE", # given as "regional office and medical center"
    "459": "VA Pacific Islands Health Care System", # Tripler (and used for JLV testing)
    "463": "ANCHORAGE AK", # given as Domiciliary - REGION 2
    "506": "ANN ARBOR", # given as medical center
    "523": "BOSTON", # given as "medical center"
    "528": "BUFFALO", # given as "medical center"
    "531": "BOISE ID", # REGION 2
    "554": "DENVER", # given as "medical center"
    "570": "FRESNO", # given as medical center
    "573": "GAINESVILLE", # given as "medical center"
    "580": "HOUSTON", # given as "medical center"
    "583": "INDIANAPOLIS", # given as medical center
    "589": "KANSAS CITY",
    "600": "LONG BEACH",
    "612": "MARTINEZ (NORTHERN CA)", # https://www.northerncalifornia.va.gov/locations/Martinez_OPC.asp - given as Medical Center
    "636": "OMAHA", # given as "medical center"
    "640": "PALO ALTO", # given as "medical center"
    "644": "PHOENIX", # VAMC
    "648": "PORTLAND OR", # given as "medical center" - REGION 2
    "653": "ROSEBURG OR", # REGION 2
    "660": "SALT LAKE CITY", # given as "medical center"
    "662": "SAN FRANCISCO",
    "663": "PUGET SOUND", # America Lake + Seattle Division - given as "medical center" Seattle - REGION 2 
    "664": "SAN DIEGO",
    "668": "SPOKANE", # Mann-Grandstaf - https://www.spokane.va.gov/ - given as "medical center" - REGION 2
    "671": "SAN ANTONIO", # given as medical center
    "678": "SOUTHERN ARIZONA",
    "687": "WALLA WALLA", # given as "medical center" - REGION 2
    "691": "WEST LA",
    "692": "WHITE CITY OR", # REGION 2
    
    "740": "VA Texas Valley Coastal Bend Health Care System", # VISN 17 - part of lift and
}

SUBSSNS = {
    "663": "Seattle",
    "663A4": "America Lake (Tacoma)", # given as DOMICILIARIES, REHABILITATION TREATMENT PROGRAM
    "663GA": "Richland", 
    "663GB": "Bremerton",
    "663GC": "Longview",
    "663GD": "Chehalis",
    
    "442": "Cheyenne",
    "442GC": "Fort Collins",
    
    "200": "VA AAC", # see more 200's in PDF p68 ie/ Austin Automation Center
    "200AA": "VA Records Center and Vault",
    "200BR": "Blind Rehabilitation System"
}

REMOTE_CHECK = ["442", "640"]
INTERESTING_CHECK = ["442", "523", "531", "640", "648", "653", "663", "668", "687", "692"]

"""
CRUDE 640 - IP VPR

	Counter({u'10.206.18.34': 4243, u'10.206.18.33': 4183, u'10.206.18.32': 4123, u'10.206.18.35': 4066, u'10.206.18.17': 3703, u'10.206.18.16': 3671, u'10.206.18.19': 3641, u'10.206.18.18': 3593, u'10.206.18.20': 3340, u'10.227.224.40': 1662, u'10.227.224.37': 1564, u'10.227.224.35': 1556, u'10.227.224.36': 1496, u'10.227.224.126': 1140, u'10.227.224.85': 1128, u'10.227.224.108': 1101, u'10.227.224.106': 1018, u'10.227.224.138': 945, u'10.206.18.70': 509, u'10.206.18.63': 492, u'10.206.18.64': 485, u'10.206.18.66': 460, u'10.206.18.69': 460, u'10.206.18.71': 438, u'10.206.18.65': 427, u'10.206.18.68': 427, u'10.206.18.90': 426, u'10.206.18.67': 425, u'10.206.18.89': 393, u'10.224.18.18': 247, u'10.224.18.17': 240, u'10.224.18.30': 236, u'10.224.18.19': 236, u'10.224.18.20': 216, u'10.224.18.31': 211, u'10.224.18.33': 207, u'10.224.18.16': 200, u'10.224.18.32': 199, u'10.206.184.44': 128, u'10.206.184.46': 126, u'10.206.184.38': 117, u'10.206.184.45': 114, u'10.206.184.67': 53, u'10.206.184.69': 48, u'10.206.184.21': 43, u'10.206.184.83': 36, u'10.227.224.167': 34, u'10.206.184.70': 33, u'10.227.224.171': 25, u'10.227.224.175': 21})
"""

# JLV_PRODUCTION_VDS_IP_ADDRESSES 
JLV_INVA_IPS = [
    '10.224.18.16',
    '10.224.18.17',
    '10.224.18.18',
    '10.224.18.19',
    '10.224.18.20',
    '10.224.18.30',
    '10.224.18.31',
    '10.224.18.32',
    '10.224.18.33',
    '10.224.18.65',
    '10.224.18.66',
    '10.224.18.67',
    '10.224.18.68',
    '10.224.18.69',
    '10.224.18.70',
    '10.224.18.71',
    '10.224.18.72',
    '10.224.18.73',
    '10.224.18.74',
    '10.224.18.75',
    '10.224.18.137',
    '10.224.18.138',
    '10.224.18.139',
    '10.224.18.140',
    '10.224.18.141',
    
    '10.206.18.16',
    '10.206.18.17',
    '10.206.18.18',
    '10.206.18.19',
    '10.206.18.20',
    '10.206.18.32',
    '10.206.18.33',
    '10.206.18.34',
    '10.206.18.35',
    '10.206.18.63',
    '10.206.18.64',
    '10.206.18.65',
    '10.206.18.66',
    '10.206.18.67',
    '10.206.18.68',
    '10.206.18.69',
    '10.206.18.70',
    '10.206.18.71',
    '10.206.18.72',
    '10.206.18.73'
]
    
JLV_CV_IPS = [    # Community Viewer - does this make sense?
    '10.227.224.167',
    '10.227.224.171',
    '10.227.224.173',
    '10.227.224.174',
    '10.227.224.175',
]

# EXTRA IPs - have 10WS too and right SMOs (with VPR) in user
JLV_INVA_EXTRA_IPS = [

	'10.206.18.89', # extra two - or fill in 72-88 too?
	'10.206.18.90',
	
	'10.206.184.21', # extra 206 for JLV?
	'10.206.184.38',
	'10.206.184.44',
	'10.206.184.45',
	'10.206.184.46',
	'10.206.184.67',
	'10.206.184.69',
	'10.206.184.70',
	'10.206.184.83',
	
	'10.227.224.35', # PERHAPS more COMMUNITY? and see above, have VPR!
	'10.227.224.36',
	'10.227.224.37',
	'10.227.224.40',
	'10.227.224.85',
    '10.227.224.106',
	'10.227.224.108',
	'10.227.224.126',
	'10.227.224.138'
	
]

JLV_ALL_IPS = list(set(JLV_INVA_IPS).union(set(JLV_INVA_EXTRA_IPS)).union(set(JLV_CV_IPS)))

# 10.184.[36|38|165].

VISTAWEBPROD_IP = '10.224.74.88'

# Question here of whether see a XWB{ old style sign on and then the real BSE one?
IMAGING_IP_BY_STATION = {
    '442': '10.152.80.125',
    '640': '10.168.30.145',
    '740': '10.141.141.40' 
}


# User identifying information from 3.081 reduction by user (vs JLV IPs)
# ... note: app is 'VISTA IMAGING VIX' and MHV probably == MyHealtheVet
# ... no special IPs as use local IPs
CVIX_MHVUSER_SSNIEN = "200:412864" # expect 2001 too and 2006_95's >> sign ons
CVIX_USER_SSNIEN = "200:217122" # expect 2001 too; 2006_95 << sign ons 

## MM == million man
MM_BOSTON_APP_SSNIEN = "523:14358"
MM_BOSTON_APP_IPS = ["10.225.31.135", "10.224.31.182"]
MM_ANN_ARBOR_APP_SSNIEN = "506:10000029117"
MM_ANN_ARBOR_APP_IPS = ["10.224.80.121", "10.225.81.65"] # first IP used in other users too

CONNECTOR_PROXIES = [

    {
        "name": "VISTALINK",
        "description": "The VistALink 1.5 resource adapter is a transport layer that provides communication between HealtheVet-VistA Java applications and VistA/M servers ... Supports VistA modules requiring this communication capability, including Patient Advocate Tracking System (PATS), Veterans Personal Finance System (VPFS) and Blind Rehabilitation",
        "notes": [
            "Kernel XUMGR key is required to create the connector proxy user account",
            "To allow VistALink access from a specific J2EE system (app server), you need an M Kernel 'connector proxy user' account. Each connector deployed on the app server uses this account to establish initial authentication and a trusted connection",
            "Hines EMC data center. These servers host the HealtheVet-VistA Blind Rehabilitation 5.0 application and require the Connector Proxy User account to execute RPC’s on your VistA/M system [ie/ this service is Hines and it gets one a/c?] ... The Blind Rehabilitation (BR) V5.0 application provides enhanced tracking, and reporting, of the blind rehabilitation services provided to veterans"
        ],
        "inMonograph": True,
        "connectionProxyNamePattern": "VISTALINK,EMC *",
        "connectionProxyExamples": "VISTALINK,EMC SSPFW SPOKANE / VISTALINK,EMC HINES",
        "rpcOptions": [
            "XOBV VISTALINK TESTER",
            "XUPS VISTALINK"
        ],
        "rpcOptionQuestions": [
            "AMOJ VISTALINK EQP INV",
            "PSN VISTALINK CONTEXT",
            "VBECS VISTALINK CONTEXT"
        ]
    },

    {
        "name": "Emergency Department Integration Software (EDIS)",
        "description": "incorporates several Web-based views that extend the current Computerized Patient Record System (CPRS) to help healthcare professionals track and manage the flow of patient care in the emergency department setting. ... EDIS runs in Adobe Flash Player",
        "notes": [
            "You must assign to users and clinical application coordinators (CACs) at least one of the following secondary menu options",
            "Site IRM will need to take note of the ACCESS and VERIFY codes and keep these codes in a place that can be later retrieved since these codes will need to be provided to the VPS point of contact in  order to configure the VETLINK VPS Kiosk",
            "Once the CONNECT,VPS user has been defined, the XMUSER menu will need to added as the PRIMARY MENU option AND add the VPS KIOSK INTERFACE broker menu as a Secondary Menu option"
        ],
        "inMonograph": True,
        "connectionProxyName": "CONNECTOR,EDIS",
        "TODO": "fill out all EDPF options",
        "rpcOptions": [
            "EDPF TRACKING MENU ALL",
            "EDPF TRACKING MENU CLINICAN"
        ]
    },

    {
        "name": "VETLINK VA Point of Service (VPS)",
        "description": "VHA has recently identified interactive kiosks as an innovation that will enable VA Medical Centers (VAMCs) to enhance services to veterans and improve the efficiency of operations. ... The VETLINK VPS Kiosk Application Server uses the RPC BROKER to make calls to the Remote Procedures (RPCs) residing on the VistA host.",
        "questions": [
            "Most GET RPCs but can write questionaires and vitals (what files?)",
            "RPC 'VPS WRITE KIOSK PARAMETERS' - what file is this?"
        ],
        "seeAlso": ["https://github.com/OSEHRA/VistA/blob/master/Packages/VA Point of Service/Patches/VPS_1.0_1/VPS-1_SEQ-1_PAT-1.TXT"],
        "inMonograph": True,
        "connectionProxyName": "CONNECT,VPS",
        "rpcOptions": ["VPS KIOSK INTERFACE"]
    },
    
    {
        "name": "DATABRIDGE,PHILIPS",
        "?": "DSS Clinical Information System (CIS)-DataBridge is a software interface solution that filters patient data information from Veterans Health Information Systems and Technology Architecture (VistA) and Clinical Patient Record System (CPRS) to a commercial Intensive Care Unit (ICU)/Anesthesia record keeping (ARK) system."
    }
    
]



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import json
import re
from collections import defaultdict, Counter

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.sankey import Sankey
import pandas as pd
try:
    import seaborn as sns
except:
    print("Ned to pip install 'seaborn'")
    raise SystemExit("Missing Required Package")
    
from visualize import flushPlots
    
"""
1. sns.barplot(x = "smoker", y = "total_bill", data =tips) ... do for IFC type
and numbers ... yes, no simple
"""
def visualizeTwo():
    ifcSplit = [
        {"type": "Placer", "total": 9999}
        
    ]
    ifcSplit = sns.load_dataset("ifcSplit")
    sns.barplot(x = "type", y = "total", data = ifcSplit)
    plt.show()
    
"""
Want side by side for User directory and steady users

Note: raw data was per day

fig, axes = plt.subplots(2,2)
axes[0, 0].hist(df['data science'])
axes[0, 1].scatter(df['Mes'], df['data science'])
axes[1, 0].plot(df['Mes'], df['machine learning'])
axes[1, 1].plot(df['Mes'], df['deep learning'])
"""

# MORE histogram etc (for trends) to try
"""
Histogram ---- in that article ---- could do spread of # of users per year since the 
start of the directory.

Scatter Plot of Consults in the three placer categories per 

[ie/ work this article]

ie/ three things:
- 1. index or not ... df ... ("Consults", X, Y, Z) ... cols ("Type", ... ... ...)
     ... index just means first column needn't be named out => special?
- 2. Histogram for per day spreads? or put into user first? yeh, user, ... over all the years spread of per year addition of users and patients ie/ median per year if yearly
buckets ...
- 3. Scatter for all three vs ?

AND

plt.annotate('Notice something?', xy=('2014-01-01', 30), xytext=('2006-01-01', 50), arrowprops={'facecolor':'red', 'shrink':0.05})
 <----------- adding annotations --- pointing to highlights
"""

# ####################### Seaborn ###################

"""
https://medium.com/@mukul.mschauhan/data-visualisation-using-seaborn-464b7c0e5122
        and seaborn in 
https://towardsdatascience.com/complete-guide-to-data-visualization-with-python-2dd74df12b5e

https://hackingandslacking.com/plotting-data-with-seaborn-and-pandas-d2499fdf6f01 (simple simple)

Also go into themes: plt.style.use('seaborn-darkgrid')

and

    sns.palplot(sns.color_palette("husl", 8)) ... husl is palette built into seaborn
    
and

    nyc_chart = sns.lineplot(x="day", y="temp", hue='year', data=nyc_df ).set_title('NYC Weather Over Time')

ie/ they do hues etc

[usual mix of snippets from different Medium Articles etc until settle down]

REM: builds over matplotlib already ie/ convenience ala the convenience methods here

Key Points to Try:
1. sns.barplot(x = "smoker", y = "total_bill", data =tips) ... do for IFC type
and numbers ... yes, no simple
2. sns.boxplot(x = , y =, data=) ... boxplot ... smoker yes/no vs IFCs ... ie/
per year do IFC breakdown NEXT?
3. sns.distplot(tips["total_bill"], bins=16, color="purple") ... I like this as a spread of per day IFC totals [or could be users involved?] ie/ distrib of typical 
4. See article has Day of Week (work off create etc date in reductions but just
record that => category color) 
    sns.lmplot(x = "total_bill", y = "tip", data = tips, hue="day")
   Could have day of week 
5. sns.heatmap(df.corr(), annot=True, fmt='.2f') ... to see correlation or otherwise of two or more variables ... do it for volumes of subtypes ie/ do they all take off together?
"""
        
# ############################## TEST DRIVER ################################
    
VISTAS_FNAMES = {"668": "Spokane", "663": "Puget Sound", "648": "Portland", "687": "Walla Walla"}
VISTA_NAMES = {"663": "PUG", "648": "POR", "668": "SPO", "687": "WWW"}
        
def main():

    assert sys.version_info >= (3, 6)
    
    data = [(1355, "PUG"), (680, "POR"), (615, "BOI"), (263, "SPO"), (37, "WCTY")]
    makeVISNSankey(data, 29, "Other IFCs Placed")
    return
        
if __name__ == "__main__":
    main()
